# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import ConcatenateLTA


def test_ConcatenateLTA_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_lta1=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-3,
        ),
        in_lta2=dict(
            argstr="%s",
            mandatory=True,
            position=-2,
        ),
        invert_1=dict(
            argstr="-invert1",
        ),
        invert_2=dict(
            argstr="-invert2",
        ),
        invert_out=dict(
            argstr="-invertout",
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            hash_files=False,
            keep_extension=True,
            name_source=["in_lta1"],
            name_template="%s_concat",
            position=-1,
        ),
        out_type=dict(
            argstr="-out_type %d",
        ),
        subject=dict(
            argstr="-subject %s",
        ),
        subjects_dir=dict(),
        tal_source_file=dict(
            argstr="-tal %s",
            extensions=None,
            position=-5,
            requires=["tal_template_file"],
        ),
        tal_template_file=dict(
            argstr="%s",
            extensions=None,
            position=-4,
            requires=["tal_source_file"],
        ),
    )
    inputs = ConcatenateLTA.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ConcatenateLTA_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = ConcatenateLTA.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..diffusion import dtiaverage


def test_dtiaverage_inputs():
    input_map = dict(
        DTI_double=dict(
            argstr="--DTI_double ",
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputs=dict(
            argstr="--inputs %s...",
        ),
        tensor_output=dict(
            argstr="--tensor_output %s",
            hash_files=False,
        ),
        verbose=dict(
            argstr="--verbose ",
        ),
    )
    inputs = dtiaverage.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_dtiaverage_outputs():
    output_map = dict(
        tensor_output=dict(
            extensions=None,
        ),
    )
    outputs = dtiaverage.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-

import os.path as op
import nibabel as nb
import numpy as np
from distutils.version import LooseVersion


from ... import logging
from ..base import traits, TraitedSpec, File, isdefined
from .base import (
    HAVE_DIPY,
    dipy_version,
    dipy_to_nipype_interface,
    get_dipy_workflows,
    DipyBaseInterface,
)

IFLOGGER = logging.getLogger("nipype.interface")

if HAVE_DIPY and LooseVersion(dipy_version()) >= LooseVersion("0.15"):
    from dipy.workflows import denoise, mask

    l_wkflw = get_dipy_workflows(denoise) + get_dipy_workflows(mask)
    for name, obj in l_wkflw:
        new_name = name.replace("Flow", "")
        globals()[new_name] = dipy_to_nipype_interface(new_name, obj)
    del l_wkflw

else:
    IFLOGGER.info(
        "We advise you to upgrade DIPY version. This upgrade will"
        " open access to more function"
    )


class ResampleInputSpec(TraitedSpec):
    in_file = File(
        exists=True, mandatory=True, desc="The input 4D diffusion-weighted image file"
    )
    vox_size = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        desc=(
            "specify the new voxel zooms. If no vox_size"
            " is set, then isotropic regridding will "
            "be performed, with spacing equal to the "
            "smallest current zoom."
        ),
    )
    interp = traits.Int(
        1,
        mandatory=True,
        usedefault=True,
        desc=("order of the interpolator (0 = nearest, 1 = linear, etc."),
    )


class ResampleOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class Resample(DipyBaseInterface):
    """
    An interface to reslicing diffusion datasets.
    See
    http://nipy.org/dipy/examples_built/reslice_datasets.html#example-reslice-datasets.

    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> reslice = dipy.Resample()
    >>> reslice.inputs.in_file = 'diffusion.nii'
    >>> reslice.run() # doctest: +SKIP
    """

    input_spec = ResampleInputSpec
    output_spec = ResampleOutputSpec

    def _run_interface(self, runtime):
        order = self.inputs.interp
        vox_size = None

        if isdefined(self.inputs.vox_size):
            vox_size = self.inputs.vox_size

        out_file = op.abspath(self._gen_outfilename())
        resample_proxy(
            self.inputs.in_file, order=order, new_zooms=vox_size, out_file=out_file
        )

        IFLOGGER.info("Resliced image saved as %s", out_file)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        fname, fext = op.splitext(op.basename(self.inputs.in_file))
        if fext == ".gz":
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        return op.abspath("%s_reslice%s" % (fname, fext))


class DenoiseInputSpec(TraitedSpec):
    in_file = File(
        exists=True, mandatory=True, desc="The input 4D diffusion-weighted image file"
    )
    in_mask = File(exists=True, desc="brain mask")
    noise_model = traits.Enum(
        "rician",
        "gaussian",
        mandatory=True,
        usedefault=True,
        desc=("noise distribution model"),
    )
    signal_mask = File(
        desc=("mask in which the mean signal " "will be computed"), exists=True
    )
    noise_mask = File(
        desc=("mask in which the standard deviation of noise " "will be computed"),
        exists=True,
    )
    patch_radius = traits.Int(1, usedefault=True, desc="patch radius")
    block_radius = traits.Int(5, usedefault=True, desc="block_radius")
    snr = traits.Float(desc="manually set an SNR")


class DenoiseOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class Denoise(DipyBaseInterface):
    """
    An interface to denoising diffusion datasets [Coupe2008]_.
    See
    http://nipy.org/dipy/examples_built/denoise_nlmeans.html#example-denoise-nlmeans.

    .. [Coupe2008] Coupe P et al., `An Optimized Blockwise Non Local Means
      Denoising Filter for 3D Magnetic Resonance Images
      <https://doi.org/10.1109%2FTMI.2007.906087>`_,
      IEEE Transactions on Medical Imaging, 27(4):425-441, 2008.


    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> denoise = dipy.Denoise()
    >>> denoise.inputs.in_file = 'diffusion.nii'
    >>> denoise.run() # doctest: +SKIP
    """

    input_spec = DenoiseInputSpec
    output_spec = DenoiseOutputSpec

    def _run_interface(self, runtime):
        out_file = op.abspath(self._gen_outfilename())

        settings = dict(mask=None, rician=(self.inputs.noise_model == "rician"))

        if isdefined(self.inputs.in_mask):
            settings["mask"] = np.asanyarray(nb.load(self.inputs.in_mask).dataobj)

        if isdefined(self.inputs.patch_radius):
            settings["patch_radius"] = self.inputs.patch_radius

        if isdefined(self.inputs.block_radius):
            settings["block_radius"] = self.inputs.block_radius

        snr = None
        if isdefined(self.inputs.snr):
            snr = self.inputs.snr

        signal_mask = None
        if isdefined(self.inputs.signal_mask):
            signal_mask = np.asanyarray(nb.load(self.inputs.signal_mask).dataobj)
        noise_mask = None
        if isdefined(self.inputs.noise_mask):
            noise_mask = np.asanyarray(nb.load(self.inputs.noise_mask).dataobj)

        _, s = nlmeans_proxy(
            self.inputs.in_file,
            settings,
            snr=snr,
            smask=signal_mask,
            nmask=noise_mask,
            out_file=out_file,
        )
        IFLOGGER.info("Denoised image saved as %s, estimated SNR=%s", out_file, str(s))
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        fname, fext = op.splitext(op.basename(self.inputs.in_file))
        if fext == ".gz":
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        return op.abspath("%s_denoise%s" % (fname, fext))


def resample_proxy(in_file, order=3, new_zooms=None, out_file=None):
    """
    Performs regridding of an image to set isotropic voxel sizes using dipy.
    """
    from dipy.align.reslice import reslice

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == ".gz":
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        out_file = op.abspath("./%s_reslice%s" % (fname, fext))

    img = nb.load(in_file)
    hdr = img.header.copy()
    data = img.get_fdata(dtype=np.float32)
    affine = img.affine
    im_zooms = hdr.get_zooms()[:3]

    if new_zooms is None:
        minzoom = np.array(im_zooms).min()
        new_zooms = tuple(np.ones((3,)) * minzoom)

    if np.all(im_zooms == new_zooms):
        return in_file

    data2, affine2 = reslice(data, affine, im_zooms, new_zooms, order=order)
    tmp_zooms = np.array(hdr.get_zooms())
    tmp_zooms[:3] = new_zooms[0]
    hdr.set_zooms(tuple(tmp_zooms))
    hdr.set_data_shape(data2.shape)
    hdr.set_xyzt_units("mm")
    nb.Nifti1Image(data2.astype(hdr.get_data_dtype()), affine2, hdr).to_filename(
        out_file
    )
    return out_file, new_zooms


def nlmeans_proxy(in_file, settings, snr=None, smask=None, nmask=None, out_file=None):
    """
    Uses non-local means to denoise 4D datasets
    """
    from dipy.denoise.nlmeans import nlmeans
    from scipy.ndimage.morphology import binary_erosion
    from scipy import ndimage

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == ".gz":
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext
        out_file = op.abspath("./%s_denoise%s" % (fname, fext))

    img = nb.load(in_file)
    hdr = img.header
    data = img.get_fdata()
    aff = img.affine

    if data.ndim < 4:
        data = data[..., np.newaxis]

    data = np.nan_to_num(data)

    if data.max() < 1.0e-4:
        raise RuntimeError("There is no signal in the image")

    df = 1.0
    if data.max() < 1000.0:
        df = 1000.0 / data.max()
        data *= df

    b0 = data[..., 0]

    if smask is None:
        smask = np.zeros_like(b0)
        smask[b0 > np.percentile(b0, 85.0)] = 1

    smask = binary_erosion(smask.astype(np.uint8), iterations=2).astype(np.uint8)

    if nmask is None:
        nmask = np.ones_like(b0, dtype=np.uint8)
        bmask = settings["mask"]
        if bmask is None:
            bmask = np.zeros_like(b0)
            bmask[b0 > np.percentile(b0[b0 > 0], 10)] = 1
            label_im, nb_labels = ndimage.label(bmask)
            sizes = ndimage.sum(bmask, label_im, range(nb_labels + 1))
            maxidx = np.argmax(sizes)
            bmask = np.zeros_like(b0, dtype=np.uint8)
            bmask[label_im == maxidx] = 1
        nmask[bmask > 0] = 0
    else:
        nmask = np.squeeze(nmask)
        nmask[nmask > 0.0] = 1
        nmask[nmask < 1] = 0
        nmask = nmask.astype(bool)

    nmask = binary_erosion(nmask, iterations=1).astype(np.uint8)

    den = np.zeros_like(data)

    est_snr = True
    if snr is not None:
        snr = [snr] * data.shape[-1]
        est_snr = False
    else:
        snr = []

    for i in range(data.shape[-1]):
        d = data[..., i]
        if est_snr:
            s = np.mean(d[smask > 0])
            n = np.std(d[nmask > 0])
            snr.append(s / n)

        den[..., i] = nlmeans(d, snr[i], **settings)

    den = np.squeeze(den)
    den /= df

    nb.Nifti1Image(den.astype(hdr.get_data_dtype()), aff, hdr).to_filename(out_file)
    return out_file, snr



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Check the resolving/rebasing feature of ``BasePath``s."""
from ... import base as nib
from ..traits_extension import rebase_path_traits, resolve_path_traits, Path


class _test_spec(nib.TraitedSpec):
    a = nib.File()
    b = nib.traits.Tuple(nib.File(), nib.File())
    c = nib.traits.List(nib.File())
    d = nib.traits.Either(nib.File(), nib.traits.Float())
    e = nib.OutputMultiObject(nib.File())
    ee = nib.OutputMultiObject(nib.Str)
    f = nib.traits.Dict(nib.Str, nib.File())
    g = nib.traits.Either(nib.File, nib.Str)
    h = nib.Str
    i = nib.traits.Either(nib.File, nib.traits.Tuple(nib.File, nib.traits.Int))
    j = nib.traits.Either(
        nib.File,
        nib.traits.Tuple(nib.File, nib.traits.Int),
        nib.traits.Dict(nib.Str, nib.File()),
    )
    k = nib.DictStrStr


def test_rebase_resolve_path_traits():
    """Check rebase_path_traits and resolve_path_traits and idempotence."""
    spec = _test_spec()

    v = "/some/path/f1.txt"
    a = rebase_path_traits(spec.trait("a"), v, "/some/path")
    assert a == Path("f1.txt")

    # Idempotence
    assert rebase_path_traits(spec.trait("a"), a, "/some/path") == a

    a = resolve_path_traits(spec.trait("a"), a, "/some/path")
    assert a == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("a"), a, "/some/path") == a

    a = rebase_path_traits(spec.trait("a"), v, "/some/other/path")
    assert a == Path(v)

    # Idempotence
    assert rebase_path_traits(spec.trait("a"), a, "/some/other/path") == a

    a = resolve_path_traits(spec.trait("a"), a, "/some/path")
    assert a == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("a"), a, "/some/path") == a

    v = ("/some/path/f1.txt", "/some/path/f2.txt")
    b = rebase_path_traits(spec.trait("b"), v, "/some/path")
    assert b == (Path("f1.txt"), Path("f2.txt"))

    # Idempotence
    assert rebase_path_traits(spec.trait("b"), b, "/some/path") == b

    b = resolve_path_traits(spec.trait("b"), b, "/some/path")
    assert b == (Path(v[0]), Path(v[1]))

    # Idempotence
    assert resolve_path_traits(spec.trait("b"), b, "/some/path") == b

    v = ["/some/path/f1.txt", "/some/path/f2.txt", "/some/path/f3.txt"]
    c = rebase_path_traits(spec.trait("c"), v, "/some/path")
    assert c == [Path("f1.txt"), Path("f2.txt"), Path("f3.txt")]

    # Idempotence
    assert rebase_path_traits(spec.trait("c"), c, "/some/path") == c

    c = resolve_path_traits(spec.trait("c"), c, "/some/path")
    assert c == [Path(vp) for vp in v]

    # Idempotence
    assert resolve_path_traits(spec.trait("c"), c, "/some/path") == c

    v = 2.0
    d = rebase_path_traits(spec.trait("d"), v, "/some/path")
    assert d == v

    d = resolve_path_traits(spec.trait("d"), d, "/some/path")
    assert d == v

    v = "/some/path/either.txt"
    d = rebase_path_traits(spec.trait("d"), v, "/some/path")
    assert d == Path("either.txt")

    # Idempotence
    assert rebase_path_traits(spec.trait("d"), d, "/some/path") == d

    d = resolve_path_traits(spec.trait("d"), d, "/some/path")
    assert d == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("d"), d, "/some/path") == d

    v = ["/some/path/f1.txt", "/some/path/f2.txt", "/some/path/f3.txt"]
    e = rebase_path_traits(spec.trait("e"), v, "/some/path")
    assert e == [Path("f1.txt"), Path("f2.txt"), Path("f3.txt")]

    # Idempotence
    assert rebase_path_traits(spec.trait("e"), e, "/some/path") == e

    e = resolve_path_traits(spec.trait("e"), e, "/some/path")
    assert e == [Path(vp) for vp in v]

    # Idempotence
    assert resolve_path_traits(spec.trait("e"), e, "/some/path") == e

    v = [["/some/path/f1.txt", "/some/path/f2.txt"], [["/some/path/f3.txt"]]]
    e = rebase_path_traits(spec.trait("e"), v, "/some/path")
    assert e == [[Path("f1.txt"), Path("f2.txt")], [[Path("f3.txt")]]]

    # Idempotence
    assert rebase_path_traits(spec.trait("e"), e, "/some/path") == e

    e = resolve_path_traits(spec.trait("e"), e, "/some/path")
    assert e == [
        [
            [Path(vpp) for vpp in vp] if isinstance(vp, list) else Path(vp)
            for vp in inner
        ]
        for inner in v
    ]

    # Idempotence
    assert resolve_path_traits(spec.trait("e"), e, "/some/path") == e

    # These are Str - no rebasing/resolving should happen
    v = [["/some/path/f1.txt", "/some/path/f2.txt"], [["/some/path/f3.txt"]]]
    ee = rebase_path_traits(spec.trait("ee"), v, "/some/path")
    assert ee == v

    # Idempotence
    assert rebase_path_traits(spec.trait("ee"), ee, "/some/path") == ee

    ee = resolve_path_traits(
        spec.trait("ee"), [["f1.txt", "f2.txt"], [["f3.txt"]]], "/some/path"
    )
    assert ee == [["f1.txt", "f2.txt"], [["f3.txt"]]]

    # Idempotence
    assert resolve_path_traits(spec.trait("ee"), ee, "/some/path") == ee

    v = {"1": "/some/path/f1.txt"}
    f = rebase_path_traits(spec.trait("f"), v, "/some")
    assert f == {"1": Path("path/f1.txt")}

    # Idempotence
    assert rebase_path_traits(spec.trait("f"), f, "/some") == f

    f = resolve_path_traits(spec.trait("f"), f, "/some")
    assert f == {k: Path(val) for k, val in v.items()}

    # Idempotence
    assert resolve_path_traits(spec.trait("f"), f, "/some") == f

    # Either(Str, File): passing in path-like apply manipulation
    v = "/some/path/either.txt"
    g = rebase_path_traits(spec.trait("g"), v, "/some/path")
    assert g == Path("either.txt")

    # Idempotence
    assert rebase_path_traits(spec.trait("g"), g, "/some/path") == g

    g = resolve_path_traits(spec.trait("g"), g, "/some/path")
    assert g == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("g"), g, "/some/path") == g

    g = rebase_path_traits(spec.trait("g"), v, "/some")
    assert g == Path("path/either.txt")

    # Idempotence
    assert rebase_path_traits(spec.trait("g"), g, "/some/path") == g

    g = resolve_path_traits(spec.trait("g"), g, "/some")
    assert g == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("g"), g, "/some/path") == g

    # Either(Str, File): passing str discards File
    v = "either.txt"
    g = rebase_path_traits(spec.trait("g"), v, "/some/path")
    assert g == v

    # Idempotence
    assert rebase_path_traits(spec.trait("g"), g, "/some/path") == g

    # This is a problematic case, it is impossible to know whether this
    # was meant to be a string or a file.
    # In this implementation, strings take precedence
    g = resolve_path_traits(spec.trait("g"), g, "/some/path")
    assert g == v

    # Idempotence
    assert resolve_path_traits(spec.trait("g"), g, "/some/path") == g

    v = "string"
    g = rebase_path_traits(spec.trait("g"), v, "/some")
    assert g == v

    # Idempotence
    assert rebase_path_traits(spec.trait("g"), g, "/some") == g

    # This is a problematic case, it is impossible to know whether this
    # was meant to be a string or a file.
    g = resolve_path_traits(spec.trait("g"), v, "/some")
    assert g == v

    # Idempotence
    assert resolve_path_traits(spec.trait("g"), g, "/some") == g

    g = rebase_path_traits(spec.trait("g"), v, "/some/path")
    assert g == v  # You dont want this one to be a Path

    # Idempotence
    assert rebase_path_traits(spec.trait("g"), g, "/some/path") == g

    # This is a problematic case, it is impossible to know whether this
    # was meant to be a string or a file.
    g = resolve_path_traits(spec.trait("g"), g, "/some/path")
    assert g == v  # You dont want this one to be a Path

    # Idempotence
    assert resolve_path_traits(spec.trait("g"), g, "/some/path") == g

    h = rebase_path_traits(spec.trait("h"), v, "/some/path")
    assert h == v

    # Idempotence
    assert rebase_path_traits(spec.trait("h"), h, "/some/path") == h

    h = resolve_path_traits(spec.trait("h"), h, "/some/path")
    assert h == v

    # Idempotence
    assert resolve_path_traits(spec.trait("h"), h, "/some/path") == h

    v = "/some/path/either/file.txt"
    i = rebase_path_traits(spec.trait("i"), v, "/some/path")
    assert i == Path("either/file.txt")

    # Idempotence
    assert rebase_path_traits(spec.trait("i"), i, "/some/path") == i

    i = resolve_path_traits(spec.trait("i"), i, "/some/path")
    assert i == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("i"), i, "/some/path") == i

    v = ("/some/path/either/tuple/file.txt", 2)
    i = rebase_path_traits(spec.trait("i"), v, "/some/path")
    assert i == (Path("either/tuple/file.txt"), 2)

    # Idempotence
    assert rebase_path_traits(spec.trait("i"), i, "/some/path") == i

    i = resolve_path_traits(spec.trait("i"), i, "/some/path")
    assert i == (Path(v[0]), v[1])

    # Idempotence
    assert resolve_path_traits(spec.trait("i"), i, "/some/path") == i

    v = "/some/path/either/file.txt"
    j = rebase_path_traits(spec.trait("j"), v, "/some/path")
    assert j == Path("either/file.txt")

    # Idempotence
    assert rebase_path_traits(spec.trait("j"), j, "/some/path") == j

    j = resolve_path_traits(spec.trait("j"), j, "/some/path")
    assert j == Path(v)

    # Idempotence
    assert resolve_path_traits(spec.trait("j"), j, "/some/path") == j

    v = ("/some/path/either/tuple/file.txt", 2)
    j = rebase_path_traits(
        spec.trait("j"), ("/some/path/either/tuple/file.txt", 2), "/some/path"
    )
    assert j == (Path("either/tuple/file.txt"), 2)

    # Idempotence
    assert rebase_path_traits(spec.trait("j"), j, "/some/path") == j

    j = resolve_path_traits(spec.trait("j"), j, "/some/path")
    assert j == (Path(v[0]), v[1])

    # Idempotence
    assert resolve_path_traits(spec.trait("j"), j, "/some/path") == j

    v = {"a": "/some/path/either/dict/file.txt"}
    j = rebase_path_traits(spec.trait("j"), v, "/some/path")
    assert j == {"a": Path("either/dict/file.txt")}

    # Idempotence
    assert rebase_path_traits(spec.trait("j"), j, "/some/path") == j

    j = resolve_path_traits(spec.trait("j"), j, "/some/path")
    assert j == {k: Path(val) for k, val in v.items()}

    # Idempotence
    assert resolve_path_traits(spec.trait("j"), j, "/some/path") == j

    v = {"path": "/some/path/f1.txt"}
    k = rebase_path_traits(spec.trait("k"), v, "/some/path")
    assert k == v

    # Idempotence
    assert rebase_path_traits(spec.trait("k"), k, "/some/path") == k

    k = resolve_path_traits(spec.trait("k"), k, "/some/path")
    assert k == v



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from ..utils.filemanip import fname_presuffix
from .base import SimpleInterface, TraitedSpec, BaseInterfaceInputSpec, traits, File
from .. import LooseVersion


class RescaleInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="Skull-stripped image to rescale")
    ref_file = File(exists=True, mandatory=True, desc="Skull-stripped reference image")
    invert = traits.Bool(desc="Invert contrast of rescaled image")
    percentile = traits.Range(
        low=0.0,
        high=50.0,
        value=0.0,
        usedefault=True,
        desc="Percentile to use for reference to allow "
        "for outliers - 1 indicates the 1st and "
        "99th percentiles in the input file will "
        "be mapped to the 99th and 1st percentiles "
        "in the reference; 0 indicates minima and "
        "maxima will be mapped",
    )


class RescaleOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="Rescaled image")


class Rescale(SimpleInterface):
    """Rescale an image

    Rescales the non-zero portion of ``in_file`` to match the bounds of the
    non-zero portion of ``ref_file``.
    Reference values in the input and reference images are defined by the
    ``percentile`` parameter, and the reference values in each image are
    identified and the remaining values are scaled accordingly.
    In the case of ``percentile == 0``, the reference values are the maxima
    and minima of each image.
    If the ``invert`` parameter is set, the input file is inverted prior to
    rescaling.

    Examples
    --------

    To use a high-resolution T1w image as a registration target for a T2\\*
    image, it may be useful to invert the T1w image and rescale to the T2\\*
    range.
    Using the 1st and 99th percentiles may reduce the impact of outlier
    voxels.

    >>> from nipype.interfaces.image import Rescale
    >>> invert_t1w = Rescale(invert=True)
    >>> invert_t1w.inputs.in_file = 'structural.nii'
    >>> invert_t1w.inputs.ref_file = 'functional.nii'
    >>> invert_t1w.inputs.percentile = 1.
    >>> res = invert_t1w.run()  # doctest: +SKIP

    """

    input_spec = RescaleInputSpec
    output_spec = RescaleOutputSpec

    def _run_interface(self, runtime):
        import numpy as np
        import nibabel as nb

        img = nb.load(self.inputs.in_file)
        data = img.get_fdata()
        ref_data = nb.load(self.inputs.ref_file).get_fdata()

        in_mask = data > 0
        ref_mask = ref_data > 0

        q = [self.inputs.percentile, 100.0 - self.inputs.percentile]
        in_low, in_high = np.percentile(data[in_mask], q)
        ref_low, ref_high = np.percentile(ref_data[ref_mask], q)
        scale_factor = (ref_high - ref_low) / (in_high - in_low)

        signal = in_high - data if self.inputs.invert else data - in_low
        out_data = in_mask * (signal * scale_factor + ref_low)

        suffix = "_inv" if self.inputs.invert else "_rescaled"
        out_file = fname_presuffix(
            self.inputs.in_file, suffix=suffix, newpath=runtime.cwd
        )
        img.__class__(out_data, img.affine, img.header).to_filename(out_file)

        self._results["out_file"] = out_file
        return runtime


_axes = ("RL", "AP", "SI")
_orientations = tuple(
    "".join((x[i], y[j], z[k]))
    for x in _axes
    for y in _axes
    for z in _axes
    if x != y != z != x
    for i in (0, 1)
    for j in (0, 1)
    for k in (0, 1)
)


class ReorientInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="Input image")
    orientation = traits.Enum(
        _orientations, usedefault=True, desc="Target axis orientation"
    )


class ReorientOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="Reoriented image")
    transform = File(
        exists=True, desc="Affine transform from input orientation to output"
    )


class Reorient(SimpleInterface):
    """Conform an image to a given orientation

Flips and reorder the image data array so that the axes match the
directions indicated in ``orientation``.
The default ``RAS`` orientation corresponds to the first axis being ordered
from left to right, the second axis from posterior to anterior, and the
third axis from inferior to superior.

For oblique images, the original orientation is considered to be the
closest plumb orientation.

No resampling is performed, and thus the output image is not de-obliqued
or registered to any other image or template.

The effective transform is calculated from the original affine matrix to
the reoriented affine matrix.

Examples
--------

If an image is not reoriented, the original file is not modified

.. testsetup::

    >>> def print_affine(matrix):
    ...     print(str(matrix).replace(']', ' ').replace('[', ' '))

>>> import numpy as np
>>> from nipype.interfaces.image import Reorient
>>> reorient = Reorient(orientation='LPS')
>>> reorient.inputs.in_file = 'segmentation0.nii.gz'
>>> res = reorient.run()
>>> res.outputs.out_file
'segmentation0.nii.gz'

>>> print_affine(np.loadtxt(res.outputs.transform))
1.  0.  0.  0.
0.  1.  0.  0.
0.  0.  1.  0.
0.  0.  0.  1.

>>> reorient.inputs.orientation = 'RAS'
>>> res = reorient.run()
>>> res.outputs.out_file  # doctest: +ELLIPSIS
'.../segmentation0_ras.nii.gz'

>>> print_affine(np.loadtxt(res.outputs.transform))
-1.   0.   0.  60.
 0.  -1.   0.  72.
 0.   0.   1.   0.
 0.   0.   0.   1.

.. testcleanup::

    >>> import os
    >>> os.unlink(res.outputs.out_file)
    >>> os.unlink(res.outputs.transform)

"""

    input_spec = ReorientInputSpec
    output_spec = ReorientOutputSpec

    def _run_interface(self, runtime):
        import numpy as np
        import nibabel as nb
        from nibabel.orientations import axcodes2ornt, ornt_transform, inv_ornt_aff

        fname = self.inputs.in_file
        orig_img = nb.load(fname)

        # Find transform from current (approximate) orientation to
        # target, in nibabel orientation matrix and affine forms
        orig_ornt = nb.io_orientation(orig_img.affine)
        targ_ornt = axcodes2ornt(self.inputs.orientation)
        transform = ornt_transform(orig_ornt, targ_ornt)
        affine_xfm = inv_ornt_aff(transform, orig_img.shape)

        # Check can be eliminated when minimum nibabel version >= 2.4
        if LooseVersion(nb.__version__) >= LooseVersion("2.4.0"):
            reoriented = orig_img.as_reoriented(transform)
        else:
            reoriented = _as_reoriented_backport(orig_img, transform)

        # Image may be reoriented
        if reoriented is not orig_img:
            suffix = "_" + self.inputs.orientation.lower()
            out_name = fname_presuffix(fname, suffix=suffix, newpath=runtime.cwd)
            reoriented.to_filename(out_name)
        else:
            out_name = fname

        mat_name = fname_presuffix(
            fname, suffix=".mat", newpath=runtime.cwd, use_ext=False
        )
        np.savetxt(mat_name, affine_xfm, fmt="%.08f")

        self._results["out_file"] = out_name
        self._results["transform"] = mat_name

        return runtime


def _as_reoriented_backport(img, ornt):
    """Backport of img.as_reoriented as of nibabel 2.4.0"""
    import numpy as np
    import nibabel as nb
    from nibabel.orientations import inv_ornt_aff

    if np.array_equal(ornt, [[0, 1], [1, 1], [2, 1]]):
        return img

    t_arr = nb.apply_orientation(img.dataobj, ornt)
    new_aff = img.affine.dot(inv_ornt_aff(ornt, img.shape))
    reoriented = img.__class__(t_arr, new_aff, img.header)

    if isinstance(reoriented, nb.Nifti1Pair):
        # Also apply the transform to the dim_info fields
        new_dim = [
            None if orig_dim is None else int(ornt[orig_dim, 0])
            for orig_dim in img.header.get_dim_info()
        ]

        reoriented.header.set_dim_info(*new_dim)

    return reoriented



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import To3D


def test_To3D_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        assumemosaic=dict(
            argstr="-assume_dicom_mosaic",
        ),
        datatype=dict(
            argstr="-datum %s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        filetype=dict(
            argstr="-%s",
        ),
        funcparams=dict(
            argstr="-time:zt %s alt+z2",
        ),
        in_folder=dict(
            argstr="%s/*.dcm",
            mandatory=True,
            position=-1,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            name_source=["in_folder"],
            name_template="%s",
        ),
        outputtype=dict(),
        skipoutliers=dict(
            argstr="-skip_outliers",
        ),
    )
    inputs = To3D.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_To3D_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = To3D.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import RobustRegister


def test_RobustRegister_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        auto_sens=dict(
            argstr="--satit",
            mandatory=True,
            xor=["outlier_sens"],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        est_int_scale=dict(
            argstr="--iscale",
        ),
        force_double=dict(
            argstr="--doubleprec",
        ),
        force_float=dict(
            argstr="--floattype",
        ),
        half_source=dict(
            argstr="--halfmov %s",
        ),
        half_source_xfm=dict(
            argstr="--halfmovlta %s",
        ),
        half_targ=dict(
            argstr="--halfdst %s",
        ),
        half_targ_xfm=dict(
            argstr="--halfdstlta %s",
        ),
        half_weights=dict(
            argstr="--halfweights %s",
        ),
        high_iterations=dict(
            argstr="--highit %d",
        ),
        in_xfm_file=dict(
            argstr="--transform",
            extensions=None,
        ),
        init_orient=dict(
            argstr="--initorient",
        ),
        iteration_thresh=dict(
            argstr="--epsit %.3f",
        ),
        least_squares=dict(
            argstr="--leastsquares",
        ),
        mask_source=dict(
            argstr="--maskmov %s",
            extensions=None,
        ),
        mask_target=dict(
            argstr="--maskdst %s",
            extensions=None,
        ),
        max_iterations=dict(
            argstr="--maxit %d",
        ),
        no_init=dict(
            argstr="--noinit",
        ),
        no_multi=dict(
            argstr="--nomulti",
        ),
        out_reg_file=dict(
            argstr="--lta %s",
            usedefault=True,
        ),
        outlier_limit=dict(
            argstr="--wlimit %.3f",
        ),
        outlier_sens=dict(
            argstr="--sat %.4f",
            mandatory=True,
            xor=["auto_sens"],
        ),
        registered_file=dict(
            argstr="--warp %s",
        ),
        source_file=dict(
            argstr="--mov %s",
            extensions=None,
            mandatory=True,
        ),
        subjects_dir=dict(),
        subsample_thresh=dict(
            argstr="--subsample %d",
        ),
        target_file=dict(
            argstr="--dst %s",
            extensions=None,
            mandatory=True,
        ),
        trans_only=dict(
            argstr="--transonly",
        ),
        weights_file=dict(
            argstr="--weights %s",
        ),
        write_vo2vox=dict(
            argstr="--vox2vox",
        ),
    )
    inputs = RobustRegister.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_RobustRegister_outputs():
    output_map = dict(
        half_source=dict(
            extensions=None,
        ),
        half_source_xfm=dict(
            extensions=None,
        ),
        half_targ=dict(
            extensions=None,
        ),
        half_targ_xfm=dict(
            extensions=None,
        ),
        half_weights=dict(
            extensions=None,
        ),
        out_reg_file=dict(
            extensions=None,
        ),
        registered_file=dict(
            extensions=None,
        ),
        weights_file=dict(
            extensions=None,
        ),
    )
    outputs = RobustRegister.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..nbs import NetworkBasedStatistic


def test_NetworkBasedStatistic_inputs():
    input_map = dict(
        edge_key=dict(
            usedefault=True,
        ),
        in_group1=dict(
            mandatory=True,
        ),
        in_group2=dict(
            mandatory=True,
        ),
        node_position_network=dict(
            extensions=None,
        ),
        number_of_permutations=dict(
            usedefault=True,
        ),
        out_nbs_network=dict(
            extensions=None,
        ),
        out_nbs_pval_network=dict(
            extensions=None,
        ),
        t_tail=dict(
            usedefault=True,
        ),
        threshold=dict(
            usedefault=True,
        ),
    )
    inputs = NetworkBasedStatistic.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_NetworkBasedStatistic_outputs():
    output_map = dict(
        nbs_network=dict(
            extensions=None,
        ),
        nbs_pval_network=dict(
            extensions=None,
        ),
        network_files=dict(),
    )
    outputs = NetworkBasedStatistic.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import Label2Vol


def test_Label2Vol_inputs():
    input_map = dict(
        annot_file=dict(
            argstr="--annot %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            requires=("subject_id", "hemi"),
            xor=("label_file", "annot_file", "seg_file", "aparc_aseg"),
        ),
        aparc_aseg=dict(
            argstr="--aparc+aseg",
            mandatory=True,
            xor=("label_file", "annot_file", "seg_file", "aparc_aseg"),
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fill_thresh=dict(
            argstr="--fillthresh %g",
        ),
        hemi=dict(
            argstr="--hemi %s",
        ),
        identity=dict(
            argstr="--identity",
            xor=("reg_file", "reg_header", "identity"),
        ),
        invert_mtx=dict(
            argstr="--invertmtx",
        ),
        label_file=dict(
            argstr="--label %s...",
            copyfile=False,
            mandatory=True,
            xor=("label_file", "annot_file", "seg_file", "aparc_aseg"),
        ),
        label_hit_file=dict(
            argstr="--hits %s",
            extensions=None,
        ),
        label_voxel_volume=dict(
            argstr="--labvoxvol %f",
        ),
        map_label_stat=dict(
            argstr="--label-stat %s",
            extensions=None,
        ),
        native_vox2ras=dict(
            argstr="--native-vox2ras",
        ),
        proj=dict(
            argstr="--proj %s %f %f %f",
            requires=("subject_id", "hemi"),
        ),
        reg_file=dict(
            argstr="--reg %s",
            extensions=None,
            xor=("reg_file", "reg_header", "identity"),
        ),
        reg_header=dict(
            argstr="--regheader %s",
            extensions=None,
            xor=("reg_file", "reg_header", "identity"),
        ),
        seg_file=dict(
            argstr="--seg %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
            xor=("label_file", "annot_file", "seg_file", "aparc_aseg"),
        ),
        subject_id=dict(
            argstr="--subject %s",
        ),
        subjects_dir=dict(),
        surface=dict(
            argstr="--surf %s",
        ),
        template_file=dict(
            argstr="--temp %s",
            extensions=None,
            mandatory=True,
        ),
        vol_label_file=dict(
            argstr="--o %s",
            extensions=None,
            genfile=True,
        ),
    )
    inputs = Label2Vol.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Label2Vol_outputs():
    output_map = dict(
        vol_label_file=dict(
            extensions=None,
        ),
    )
    outputs = Label2Vol.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import Allineate


def test_Allineate_inputs():
    input_map = dict(
        allcostx=dict(
            argstr="-allcostx |& tee %s",
            extensions=None,
            position=-1,
            xor=["out_file", "out_matrix", "out_param_file", "out_weight_file"],
        ),
        args=dict(
            argstr="%s",
        ),
        autobox=dict(
            argstr="-autobox",
        ),
        automask=dict(
            argstr="-automask+%d",
        ),
        autoweight=dict(
            argstr="-autoweight%s",
        ),
        center_of_mass=dict(
            argstr="-cmass%s",
        ),
        check=dict(
            argstr="-check %s",
        ),
        convergence=dict(
            argstr="-conv %f",
        ),
        cost=dict(
            argstr="-cost %s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        epi=dict(
            argstr="-EPI",
        ),
        final_interpolation=dict(
            argstr="-final %s",
        ),
        fine_blur=dict(
            argstr="-fineblur %f",
        ),
        in_file=dict(
            argstr="-source %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        in_matrix=dict(
            argstr="-1Dmatrix_apply %s",
            extensions=None,
            position=-3,
            xor=["out_matrix"],
        ),
        in_param_file=dict(
            argstr="-1Dparam_apply %s",
            extensions=None,
            xor=["out_param_file"],
        ),
        interpolation=dict(
            argstr="-interp %s",
        ),
        master=dict(
            argstr="-master %s",
            extensions=None,
        ),
        maxrot=dict(
            argstr="-maxrot %f",
        ),
        maxscl=dict(
            argstr="-maxscl %f",
        ),
        maxshf=dict(
            argstr="-maxshf %f",
        ),
        maxshr=dict(
            argstr="-maxshr %f",
        ),
        newgrid=dict(
            argstr="-newgrid %f",
        ),
        nmatch=dict(
            argstr="-nmatch %d",
        ),
        no_pad=dict(
            argstr="-nopad",
        ),
        nomask=dict(
            argstr="-nomask",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        nwarp=dict(
            argstr="-nwarp %s",
        ),
        nwarp_fixdep=dict(
            argstr="-nwarp_fixdep%s...",
        ),
        nwarp_fixmot=dict(
            argstr="-nwarp_fixmot%s...",
        ),
        one_pass=dict(
            argstr="-onepass",
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            hash_files=False,
            name_source="in_file",
            name_template="%s_allineate",
            xor=["allcostx"],
        ),
        out_matrix=dict(
            argstr="-1Dmatrix_save %s",
            extensions=None,
            xor=["in_matrix", "allcostx"],
        ),
        out_param_file=dict(
            argstr="-1Dparam_save %s",
            extensions=None,
            xor=["in_param_file", "allcostx"],
        ),
        out_weight_file=dict(
            argstr="-wtprefix %s",
            extensions=None,
            xor=["allcostx"],
        ),
        outputtype=dict(),
        overwrite=dict(
            argstr="-overwrite",
        ),
        quiet=dict(
            argstr="-quiet",
        ),
        reference=dict(
            argstr="-base %s",
            extensions=None,
        ),
        replacebase=dict(
            argstr="-replacebase",
        ),
        replacemeth=dict(
            argstr="-replacemeth %s",
        ),
        source_automask=dict(
            argstr="-source_automask+%d",
        ),
        source_mask=dict(
            argstr="-source_mask %s",
            extensions=None,
        ),
        two_best=dict(
            argstr="-twobest %d",
        ),
        two_blur=dict(
            argstr="-twoblur %f",
        ),
        two_first=dict(
            argstr="-twofirst",
        ),
        two_pass=dict(
            argstr="-twopass",
        ),
        usetemp=dict(
            argstr="-usetemp",
        ),
        verbose=dict(
            argstr="-verb",
        ),
        warp_type=dict(
            argstr="-warp %s",
        ),
        warpfreeze=dict(
            argstr="-warpfreeze",
        ),
        weight=dict(
            argstr="-weight %s",
        ),
        weight_file=dict(
            argstr="-weight %s",
            deprecated="1.0.0",
            extensions=None,
            new_name="weight",
        ),
        zclip=dict(
            argstr="-zclip",
        ),
    )
    inputs = Allineate.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Allineate_outputs():
    output_map = dict(
        allcostx=dict(
            extensions=None,
        ),
        out_file=dict(
            extensions=None,
        ),
        out_matrix=dict(
            extensions=None,
        ),
        out_param_file=dict(
            extensions=None,
        ),
        out_weight_file=dict(
            extensions=None,
        ),
    )
    outputs = Allineate.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The testing directory contains a small set of imaging files to be
used for doctests only.
"""

import os

# Discover directory path
filepath = os.path.abspath(__file__)
basedir = os.path.dirname(filepath)

funcfile = os.path.join(basedir, "data", "functional.nii")
anatfile = os.path.join(basedir, "data", "structural.nii")
template = funcfile
transfm = funcfile

from . import decorators
from .utils import package_check, TempFATFS

skipif = decorators.dec.skipif


def example_data(infile="functional.nii"):
    """returns path to empty example data files for doc tests
    it will raise an exception if filename is not in the directory"""

    filepath = os.path.abspath(__file__)
    basedir = os.path.dirname(filepath)
    outfile = os.path.join(basedir, "data", infile)
    if not os.path.exists(outfile):
        raise IOError("%s empty data file does NOT exist" % outfile)

    return outfile



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for the engine utils module
"""
import os
from copy import deepcopy
import pytest

from ... import engine as pe
from ....interfaces import base as nib
from ....interfaces import utility as niu
from .... import config
from ..utils import (
    clean_working_directory,
    write_workflow_prov,
    load_resultfile,
    format_node,
)


class InputSpec(nib.TraitedSpec):
    in_file = nib.File(exists=True, copyfile=True)


class OutputSpec(nib.TraitedSpec):
    output1 = nib.traits.List(nib.traits.Int, desc="outputs")


class UtilsTestInterface(nib.BaseInterface):
    input_spec = InputSpec
    output_spec = OutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["output1"] = [1]
        return outputs


def test_identitynode_removal(tmpdir):
    def test_function(arg1, arg2, arg3):
        import numpy as np

        return (np.array(arg1) + arg2 + arg3).tolist()

    wf = pe.Workflow(name="testidentity", base_dir=tmpdir.strpath)

    n1 = pe.Node(
        niu.IdentityInterface(fields=["a", "b"]), name="src", base_dir=tmpdir.strpath
    )
    n1.iterables = ("b", [0, 1, 2, 3])
    n1.inputs.a = [0, 1, 2, 3]

    n2 = pe.Node(niu.Select(), name="selector", base_dir=tmpdir.strpath)
    wf.connect(n1, ("a", test_function, 1, -1), n2, "inlist")
    wf.connect(n1, "b", n2, "index")

    n3 = pe.Node(
        niu.IdentityInterface(fields=["c", "d"]), name="passer", base_dir=tmpdir.strpath
    )
    n3.inputs.c = [1, 2, 3, 4]
    wf.connect(n2, "out", n3, "d")

    n4 = pe.Node(niu.Select(), name="selector2", base_dir=tmpdir.strpath)
    wf.connect(n3, ("c", test_function, 1, -1), n4, "inlist")
    wf.connect(n3, "d", n4, "index")

    fg = wf._create_flat_graph()
    wf._set_needed_outputs(fg)
    eg = pe.generate_expanded_graph(deepcopy(fg))
    assert len(eg.nodes()) == 8


def test_clean_working_directory(tmpdir):
    class OutputSpec(nib.TraitedSpec):
        files = nib.traits.List(nib.File)
        others = nib.File()

    class InputSpec(nib.TraitedSpec):
        infile = nib.File()

    outputs = OutputSpec()
    inputs = InputSpec()

    filenames = [
        "file.hdr",
        "file.img",
        "file.BRIK",
        "file.HEAD",
        "_0x1234.json",
        "foo.txt",
    ]
    outfiles = []
    for filename in filenames:
        outfile = tmpdir.join(filename)
        outfile.write("dummy")
        outfiles.append(outfile.strpath)
    outputs.files = outfiles[:4:2]
    outputs.others = outfiles[5]
    inputs.infile = outfiles[-1]
    needed_outputs = ["files"]
    config.set_default_config()
    assert os.path.exists(outfiles[5])
    config.set_default_config()
    config.set("execution", "remove_unnecessary_outputs", False)
    out = clean_working_directory(
        outputs, tmpdir.strpath, inputs, needed_outputs, deepcopy(config._sections)
    )
    assert os.path.exists(outfiles[5])
    assert out.others == outfiles[5]
    config.set("execution", "remove_unnecessary_outputs", True)
    out = clean_working_directory(
        outputs, tmpdir.strpath, inputs, needed_outputs, deepcopy(config._sections)
    )
    assert os.path.exists(outfiles[1])
    assert os.path.exists(outfiles[3])
    assert os.path.exists(outfiles[4])
    assert not os.path.exists(outfiles[5])
    assert out.others == nib.Undefined
    assert len(out.files) == 2
    config.set_default_config()


def create_wf(name):
    """Creates a workflow for the following tests"""

    def fwhm(fwhm):
        return fwhm

    pipe = pe.Workflow(name=name)
    process = pe.Node(
        niu.Function(input_names=["fwhm"], output_names=["fwhm"], function=fwhm),
        name="proc",
    )
    process.iterables = ("fwhm", [0])
    process2 = pe.Node(
        niu.Function(input_names=["fwhm"], output_names=["fwhm"], function=fwhm),
        name="proc2",
    )
    process2.iterables = ("fwhm", [0])
    pipe.connect(process, "fwhm", process2, "fwhm")
    return pipe


def test_multi_disconnected_iterable(tmpdir):
    metawf = pe.Workflow(name="meta")
    metawf.base_dir = tmpdir.strpath
    metawf.add_nodes([create_wf("wf%d" % i) for i in range(30)])
    eg = metawf.run(plugin="Linear")
    assert len(eg.nodes()) == 60


def test_provenance(tmpdir):
    metawf = pe.Workflow(name="meta")
    metawf.base_dir = tmpdir.strpath
    metawf.add_nodes([create_wf("wf%d" % i) for i in range(1)])
    eg = metawf.run(plugin="Linear")
    prov_base = tmpdir.join("workflow_provenance_test").strpath
    psg = write_workflow_prov(eg, prov_base, format="all")
    assert len(psg.bundles) == 2
    assert len(psg.get_records()) == 7


def dummy_func(value):
    return value + 1


def test_mapnode_crash(tmpdir):
    """Test mapnode crash when stop_on_first_crash is True"""
    cwd = os.getcwd()
    node = pe.MapNode(
        niu.Function(
            input_names=["WRONG"], output_names=["newstring"], function=dummy_func
        ),
        iterfield=["WRONG"],
        name="myfunc",
    )
    node.inputs.WRONG = ["string{}".format(i) for i in range(3)]
    node.config = deepcopy(config._sections)
    node.config["execution"]["stop_on_first_crash"] = True
    node.base_dir = tmpdir.strpath
    with pytest.raises(TypeError):
        node.run()
    os.chdir(cwd)


def test_mapnode_crash2(tmpdir):
    """Test mapnode crash when stop_on_first_crash is False"""
    cwd = os.getcwd()
    node = pe.MapNode(
        niu.Function(
            input_names=["WRONG"], output_names=["newstring"], function=dummy_func
        ),
        iterfield=["WRONG"],
        name="myfunc",
    )
    node.inputs.WRONG = ["string{}".format(i) for i in range(3)]
    node.base_dir = tmpdir.strpath

    with pytest.raises(Exception):
        node.run()
    os.chdir(cwd)


def test_mapnode_crash3(tmpdir):
    """Test mapnode crash when mapnode is embedded in a workflow"""
    tmpdir.chdir()
    node = pe.MapNode(
        niu.Function(
            input_names=["WRONG"], output_names=["newstring"], function=dummy_func
        ),
        iterfield=["WRONG"],
        name="myfunc",
    )
    node.inputs.WRONG = ["string{}".format(i) for i in range(3)]
    wf = pe.Workflow("testmapnodecrash")
    wf.add_nodes([node])
    wf.base_dir = tmpdir.strpath
    # changing crashdump dir to current working directory (to avoid problems with read-only systems)
    wf.config["execution"]["crashdump_dir"] = os.getcwd()
    with pytest.raises(RuntimeError):
        wf.run(plugin="Linear")


class StrPathConfuserInputSpec(nib.TraitedSpec):
    in_str = nib.traits.String()


class StrPathConfuserOutputSpec(nib.TraitedSpec):
    out_tuple = nib.traits.Tuple(nib.File, nib.traits.String)
    out_dict_path = nib.traits.Dict(nib.traits.String, nib.File(exists=True))
    out_dict_str = nib.traits.DictStrStr()
    out_list = nib.traits.List(nib.traits.String)
    out_str = nib.traits.String()
    out_path = nib.File(exists=True)


class StrPathConfuser(nib.SimpleInterface):
    input_spec = StrPathConfuserInputSpec
    output_spec = StrPathConfuserOutputSpec

    def _run_interface(self, runtime):
        out_path = os.path.abspath(os.path.basename(self.inputs.in_str) + "_path")
        open(out_path, "w").close()
        self._results["out_str"] = self.inputs.in_str
        self._results["out_path"] = out_path
        self._results["out_tuple"] = (out_path, self.inputs.in_str)
        self._results["out_dict_path"] = {self.inputs.in_str: out_path}
        self._results["out_dict_str"] = {self.inputs.in_str: self.inputs.in_str}
        self._results["out_list"] = [self.inputs.in_str] * 2
        return runtime


def test_modify_paths_bug(tmpdir):
    """
    There was a bug in which, if the current working directory contained a file with the name
    of an output String, the string would get transformed into a path, and generally wreak havoc.
    This attempts to replicate that condition, using an object with strings and paths in various
    trait configurations, to ensure that the guards added resolve the issue.
    Please see https://github.com/nipy/nipype/issues/2944 for more details.
    """
    tmpdir.chdir()

    spc = pe.Node(StrPathConfuser(in_str="2"), name="spc")

    open("2", "w").close()

    outputs = spc.run().outputs

    # Basic check that string was not manipulated
    out_str = outputs.out_str
    assert out_str == "2"

    # Check path exists and is absolute
    out_path = outputs.out_path
    assert os.path.isabs(out_path)

    # Assert data structures pass through correctly
    assert outputs.out_tuple == (out_path, out_str)
    assert outputs.out_dict_path == {out_str: out_path}
    assert outputs.out_dict_str == {out_str: out_str}
    assert outputs.out_list == [out_str] * 2


@pytest.mark.parametrize("use_relative", [True, False])
def test_save_load_resultfile(tmpdir, use_relative):
    """Test minimally the save/load functions for result files."""
    from shutil import copytree, rmtree

    tmpdir.chdir()

    old_use_relative = config.getboolean("execution", "use_relative_paths")
    config.set("execution", "use_relative_paths", use_relative)

    spc = pe.Node(StrPathConfuser(in_str="2"), name="spc")
    spc.base_dir = tmpdir.mkdir("node").strpath

    result = spc.run()

    loaded_result = load_resultfile(
        tmpdir.join("node").join("spc").join("result_spc.pklz").strpath
    )

    assert result.runtime.dictcopy() == loaded_result.runtime.dictcopy()
    assert result.inputs == loaded_result.inputs
    assert result.outputs.get() == loaded_result.outputs.get()

    # Test the mobility of the result file.
    copytree(tmpdir.join("node").strpath, tmpdir.join("node2").strpath)
    rmtree(tmpdir.join("node").strpath)

    if use_relative:
        loaded_result2 = load_resultfile(
            tmpdir.join("node2").join("spc").join("result_spc.pklz").strpath
        )

        assert result.runtime.dictcopy() == loaded_result2.runtime.dictcopy()
        assert result.inputs == loaded_result2.inputs
        assert loaded_result2.outputs.get() != result.outputs.get()
        newpath = result.outputs.out_path.replace("/node/", "/node2/")
        assert loaded_result2.outputs.out_path == newpath
        assert loaded_result2.outputs.out_tuple[0] == newpath
        assert loaded_result2.outputs.out_dict_path["2"] == newpath
    else:
        with pytest.raises(nib.TraitError):
            load_resultfile(
                tmpdir.join("node2").join("spc").join("result_spc.pklz").strpath
            )

    config.set("execution", "use_relative_paths", old_use_relative)


def test_format_node():
    node = pe.Node(niu.IdentityInterface(fields=["a", "b"]), name="node")
    serialized = format_node(node)
    workspace = {"Node": pe.Node}
    exec("\n".join(serialized), workspace)
    assert workspace["node"].interface._fields == node.interface._fields



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..odf import ODFRecon


def test_ODFRecon_inputs():
    input_map = dict(
        DWI=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        args=dict(
            argstr="%s",
        ),
        dsi=dict(
            argstr="-dsi",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        filter=dict(
            argstr="-f",
        ),
        image_orientation_vectors=dict(
            argstr="-iop %f",
        ),
        matrix=dict(
            argstr="-mat %s",
            extensions=None,
            mandatory=True,
        ),
        n_b0=dict(
            argstr="-b0 %s",
            mandatory=True,
        ),
        n_directions=dict(
            argstr="%s",
            mandatory=True,
            position=2,
        ),
        n_output_directions=dict(
            argstr="%s",
            mandatory=True,
            position=3,
        ),
        oblique_correction=dict(
            argstr="-oc",
        ),
        out_prefix=dict(
            argstr="%s",
            position=4,
            usedefault=True,
        ),
        output_entropy=dict(
            argstr="-oe",
        ),
        output_type=dict(
            argstr="-ot %s",
            usedefault=True,
        ),
        sharpness=dict(
            argstr="-s %f",
        ),
        subtract_background=dict(
            argstr="-bg",
        ),
    )
    inputs = ODFRecon.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ODFRecon_outputs():
    output_map = dict(
        B0=dict(
            extensions=None,
        ),
        DWI=dict(
            extensions=None,
        ),
        ODF=dict(
            extensions=None,
        ),
        entropy=dict(
            extensions=None,
        ),
        max=dict(
            extensions=None,
        ),
    )
    outputs = ODFRecon.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import PRELUDE


def test_PRELUDE_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        complex_phase_file=dict(
            argstr="--complex=%s",
            extensions=None,
            mandatory=True,
            xor=["magnitude_file", "phase_file"],
        ),
        end=dict(
            argstr="--end=%d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        label_file=dict(
            argstr="--labels=%s",
            extensions=None,
            hash_files=False,
        ),
        labelprocess2d=dict(
            argstr="--labelslices",
        ),
        magnitude_file=dict(
            argstr="--abs=%s",
            extensions=None,
            mandatory=True,
            xor=["complex_phase_file"],
        ),
        mask_file=dict(
            argstr="--mask=%s",
            extensions=None,
        ),
        num_partitions=dict(
            argstr="--numphasesplit=%d",
        ),
        output_type=dict(),
        phase_file=dict(
            argstr="--phase=%s",
            extensions=None,
            mandatory=True,
            xor=["complex_phase_file"],
        ),
        process2d=dict(
            argstr="--slices",
            xor=["labelprocess2d"],
        ),
        process3d=dict(
            argstr="--force3D",
            xor=["labelprocess2d", "process2d"],
        ),
        rawphase_file=dict(
            argstr="--rawphase=%s",
            extensions=None,
            hash_files=False,
        ),
        removeramps=dict(
            argstr="--removeramps",
        ),
        savemask_file=dict(
            argstr="--savemask=%s",
            extensions=None,
            hash_files=False,
        ),
        start=dict(
            argstr="--start=%d",
        ),
        threshold=dict(
            argstr="--thresh=%.10f",
        ),
        unwrapped_phase_file=dict(
            argstr="--unwrap=%s",
            extensions=None,
            genfile=True,
            hash_files=False,
        ),
    )
    inputs = PRELUDE.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_PRELUDE_outputs():
    output_map = dict(
        unwrapped_phase_file=dict(
            extensions=None,
        ),
    )
    outputs = PRELUDE.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..developer import JistBrainMp2rageDuraEstimation


def test_JistBrainMp2rageDuraEstimation_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inDistance=dict(
            argstr="--inDistance %f",
        ),
        inSecond=dict(
            argstr="--inSecond %s",
            extensions=None,
        ),
        inSkull=dict(
            argstr="--inSkull %s",
            extensions=None,
        ),
        inoutput=dict(
            argstr="--inoutput %s",
        ),
        null=dict(
            argstr="--null %s",
        ),
        outDura=dict(
            argstr="--outDura %s",
            hash_files=False,
        ),
        xDefaultMem=dict(
            argstr="-xDefaultMem %d",
        ),
        xMaxProcess=dict(
            argstr="-xMaxProcess %d",
            usedefault=True,
        ),
        xPrefExt=dict(
            argstr="--xPrefExt %s",
        ),
    )
    inputs = JistBrainMp2rageDuraEstimation.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_JistBrainMp2rageDuraEstimation_outputs():
    output_map = dict(
        outDura=dict(
            extensions=None,
        ),
    )
    outputs = JistBrainMp2rageDuraEstimation.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import Level1Design


def test_Level1Design_inputs():
    input_map = dict(
        bases=dict(
            mandatory=True,
        ),
        contrasts=dict(),
        interscan_interval=dict(
            mandatory=True,
        ),
        model_serial_correlations=dict(
            mandatory=True,
        ),
        orthogonalization=dict(
            usedefault=True,
        ),
        session_info=dict(
            mandatory=True,
        ),
    )
    inputs = Level1Design.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Level1Design_outputs():
    output_map = dict(
        ev_files=dict(),
        fsf_files=dict(),
    )
    outputs = Level1Design.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

import os

from ...base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)


class BRAINSFitInputSpec(CommandLineInputSpec):
    fixedVolume = File(
        desc="Input fixed image (the moving image will be transformed into this image space).",
        exists=True,
        argstr="--fixedVolume %s",
    )
    movingVolume = File(
        desc="Input moving image (this image will be transformed into the fixed image space).",
        exists=True,
        argstr="--movingVolume %s",
    )
    samplingPercentage = traits.Float(
        desc="Fraction of voxels of the fixed image that will be used for registration. The number has to be larger than zero and less or equal to one. Higher values increase the computation time but may give more accurate results. You can also limit the sampling focus with ROI masks and ROIAUTO mask generation. The default is 0.002 (use approximately 0.2% of voxels, resulting in 100000 samples in a 512x512x192 volume) to provide a very fast registration in most cases. Typical values range from 0.01 (1%) for low detail images to 0.2 (20%) for high detail images.",
        argstr="--samplingPercentage %f",
    )
    splineGridSize = InputMultiPath(
        traits.Int,
        desc="Number of BSpline grid subdivisions along each axis of the fixed image, centered on the image space. Values must be 3 or higher for the BSpline to be correctly computed.",
        sep=",",
        argstr="--splineGridSize %s",
    )
    linearTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Output estimated transform - in case the computed transform is not BSpline. NOTE: You must set at least one output object (transform and/or output volume).",
        argstr="--linearTransform %s",
    )
    bsplineTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Output estimated transform - in case the computed transform is BSpline. NOTE: You must set at least one output object (transform and/or output volume).",
        argstr="--bsplineTransform %s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Output image: the moving image warped to the fixed image space. NOTE: You must set at least one output object (transform and/or output volume).",
        argstr="--outputVolume %s",
    )
    initialTransform = File(
        desc="Transform to be applied to the moving image to initialize the registration.  This can only be used if Initialize Transform Mode is Off.",
        exists=True,
        argstr="--initialTransform %s",
    )
    initializeTransformMode = traits.Enum(
        "Off",
        "useMomentsAlign",
        "useCenterOfHeadAlign",
        "useGeometryAlign",
        "useCenterOfROIAlign",
        desc="Determine how to initialize the transform center.  useMomentsAlign assumes that the center of mass of the images represent similar structures.  useCenterOfHeadAlign attempts to use the top of head and shape of neck to drive a center of mass estimate. useGeometryAlign on assumes that the center of the voxel lattice of the images represent similar structures.  Off assumes that the physical space of the images are close.  This flag is mutually exclusive with the Initialization transform.",
        argstr="--initializeTransformMode %s",
    )
    useRigid = traits.Bool(
        desc="Perform a rigid registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useRigid ",
    )
    useScaleVersor3D = traits.Bool(
        desc="Perform a ScaleVersor3D registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useScaleVersor3D ",
    )
    useScaleSkewVersor3D = traits.Bool(
        desc="Perform a ScaleSkewVersor3D registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useScaleSkewVersor3D ",
    )
    useAffine = traits.Bool(
        desc="Perform an Affine registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useAffine ",
    )
    useBSpline = traits.Bool(
        desc="Perform a BSpline registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useBSpline ",
    )
    useSyN = traits.Bool(
        desc="Perform a SyN registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useSyN ",
    )
    useComposite = traits.Bool(
        desc="Perform a Composite registration as part of the sequential registration steps.  This family of options overrides the use of transformType if any of them are set.",
        argstr="--useComposite ",
    )
    maskProcessingMode = traits.Enum(
        "NOMASK",
        "ROIAUTO",
        "ROI",
        desc="Specifies a mask to only consider a certain image region for the registration.  If ROIAUTO is chosen, then the mask is computed using Otsu thresholding and hole filling. If ROI is chosen then the mask has to be specified as in input.",
        argstr="--maskProcessingMode %s",
    )
    fixedBinaryVolume = File(
        desc="Fixed Image binary mask volume, required if Masking Option is ROI. Image areas where the mask volume has zero value are ignored during the registration.",
        exists=True,
        argstr="--fixedBinaryVolume %s",
    )
    movingBinaryVolume = File(
        desc="Moving Image binary mask volume, required if Masking Option is ROI. Image areas where the mask volume has zero value are ignored during the registration.",
        exists=True,
        argstr="--movingBinaryVolume %s",
    )
    outputFixedVolumeROI = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="ROI that is automatically computed from the fixed image. Only available if Masking Option is ROIAUTO. Image areas where the mask volume has zero value are ignored during the registration.",
        argstr="--outputFixedVolumeROI %s",
    )
    outputMovingVolumeROI = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="ROI that is automatically computed from the moving image. Only available if Masking Option is ROIAUTO. Image areas where the mask volume has zero value are ignored during the registration.",
        argstr="--outputMovingVolumeROI %s",
    )
    useROIBSpline = traits.Bool(
        desc="If enabled then the bounding box of the input ROIs defines the BSpline grid support region. Otherwise the BSpline grid support region is the whole fixed image.",
        argstr="--useROIBSpline ",
    )
    histogramMatch = traits.Bool(
        desc="Apply histogram matching operation for the input images to make them more similar.  This is suitable for images of the same modality that may have different brightness or contrast, but the same overall intensity profile. Do NOT use if registering images from different modalities.",
        argstr="--histogramMatch ",
    )
    medianFilterSize = InputMultiPath(
        traits.Int,
        desc="Apply median filtering to reduce noise in the input volumes. The 3 values specify the radius for the optional MedianImageFilter preprocessing in all 3 directions (in voxels).",
        sep=",",
        argstr="--medianFilterSize %s",
    )
    removeIntensityOutliers = traits.Float(
        desc="Remove very high and very low intensity voxels from the input volumes. The parameter specifies the half percentage to decide outliers of image intensities. The default value is zero, which means no outlier removal. If the value of 0.005 is given, the 0.005% of both tails will be thrown away, so 0.01% of intensities in total would be ignored in the statistic calculation.",
        argstr="--removeIntensityOutliers %f",
    )
    fixedVolume2 = File(
        desc="Input fixed image that will be used for multimodal registration. (the moving image will be transformed into this image space).",
        exists=True,
        argstr="--fixedVolume2 %s",
    )
    movingVolume2 = File(
        desc="Input moving image that will be used for multimodal registration(this image will be transformed into the fixed image space).",
        exists=True,
        argstr="--movingVolume2 %s",
    )
    outputVolumePixelType = traits.Enum(
        "float",
        "short",
        "ushort",
        "int",
        "uint",
        "uchar",
        desc="Data type for representing a voxel of the Output Volume.",
        argstr="--outputVolumePixelType %s",
    )
    backgroundFillValue = traits.Float(
        desc="This value will be used for filling those areas of the output image that have no corresponding voxels in the input moving image.",
        argstr="--backgroundFillValue %f",
    )
    scaleOutputValues = traits.Bool(
        desc="If true, and the voxel values do not fit within the minimum and maximum values of the desired outputVolumePixelType, then linearly scale the min/max output image voxel values to fit within the min/max range of the outputVolumePixelType.",
        argstr="--scaleOutputValues ",
    )
    interpolationMode = traits.Enum(
        "NearestNeighbor",
        "Linear",
        "ResampleInPlace",
        "BSpline",
        "WindowedSinc",
        "Hamming",
        "Cosine",
        "Welch",
        "Lanczos",
        "Blackman",
        desc="Type of interpolation to be used when applying transform to moving volume.  Options are Linear, NearestNeighbor, BSpline, WindowedSinc, Hamming, Cosine, Welch, Lanczos, or ResampleInPlace.  The ResampleInPlace option will create an image with the same discrete voxel values and will adjust the origin and direction of the physical space interpretation.",
        argstr="--interpolationMode %s",
    )
    numberOfIterations = InputMultiPath(
        traits.Int,
        desc="The maximum number of iterations to try before stopping the optimization. When using a lower value (500-1000) then the registration is forced to terminate earlier but there is a higher risk of stopping before an optimal solution is reached.",
        sep=",",
        argstr="--numberOfIterations %s",
    )
    maximumStepLength = traits.Float(
        desc="Starting step length of the optimizer. In general, higher values allow for recovering larger initial misalignments but there is an increased chance that the registration will not converge.",
        argstr="--maximumStepLength %f",
    )
    minimumStepLength = InputMultiPath(
        traits.Float,
        desc="Each step in the optimization takes steps at least this big.  When none are possible, registration is complete. Smaller values allows the optimizer to make smaller adjustments, but the registration time may increase.",
        sep=",",
        argstr="--minimumStepLength %s",
    )
    relaxationFactor = traits.Float(
        desc="Specifies how quickly the optimization step length is decreased during registration. The value must be larger than 0 and smaller than 1. Larger values result in slower step size decrease, which allow for recovering larger initial misalignments but it increases the registration time and the chance that the registration will not converge.",
        argstr="--relaxationFactor %f",
    )
    translationScale = traits.Float(
        desc="How much to scale up changes in position (in mm) compared to unit rotational changes (in radians) -- decrease this to allow for more rotation in the search pattern.",
        argstr="--translationScale %f",
    )
    reproportionScale = traits.Float(
        desc="ScaleVersor3D 'Scale' compensation factor.  Increase this to allow for more rescaling in a ScaleVersor3D or ScaleSkewVersor3D search pattern.  1.0 works well with a translationScale of 1000.0",
        argstr="--reproportionScale %f",
    )
    skewScale = traits.Float(
        desc="ScaleSkewVersor3D Skew compensation factor.  Increase this to allow for more skew in a ScaleSkewVersor3D search pattern.  1.0 works well with a translationScale of 1000.0",
        argstr="--skewScale %f",
    )
    maxBSplineDisplacement = traits.Float(
        desc="Maximum allowed displacements in image physical coordinates (mm) for BSpline control grid along each axis.  A value of 0.0 indicates that the problem should be unbounded.  NOTE:  This only constrains the BSpline portion, and does not limit the displacement from the associated bulk transform.  This can lead to a substantial reduction in computation time in the BSpline optimizer.,       ",
        argstr="--maxBSplineDisplacement %f",
    )
    fixedVolumeTimeIndex = traits.Int(
        desc="The index in the time series for the 3D fixed image to fit. Only allowed if the fixed input volume is 4-dimensional.",
        argstr="--fixedVolumeTimeIndex %d",
    )
    movingVolumeTimeIndex = traits.Int(
        desc="The index in the time series for the 3D moving image to fit. Only allowed if the moving input volume is 4-dimensional",
        argstr="--movingVolumeTimeIndex %d",
    )
    numberOfHistogramBins = traits.Int(
        desc="The number of histogram levels used for mutual information metric estimation.",
        argstr="--numberOfHistogramBins %d",
    )
    numberOfMatchPoints = traits.Int(
        desc="Number of histogram match points used for mutual information metric estimation.",
        argstr="--numberOfMatchPoints %d",
    )
    costMetric = traits.Enum(
        "MMI",
        "MSE",
        "NC",
        "MIH",
        desc="The cost metric to be used during fitting. Defaults to MMI. Options are MMI (Mattes Mutual Information), MSE (Mean Square Error), NC (Normalized Correlation), MC (Match Cardinality for binary images)",
        argstr="--costMetric %s",
    )
    maskInferiorCutOffFromCenter = traits.Float(
        desc="If Initialize Transform Mode is set to useCenterOfHeadAlign or Masking Option is ROIAUTO then this value defines the how much is cut of from the inferior part of the image. The cut-off distance is specified in millimeters, relative to the image center. If the value is 1000 or larger then no cut-off performed.",
        argstr="--maskInferiorCutOffFromCenter %f",
    )
    ROIAutoDilateSize = traits.Float(
        desc="This flag is only relevant when using ROIAUTO mode for initializing masks.  It defines the final dilation size to capture a bit of background outside the tissue region.  A setting of 10mm has been shown to help regularize a BSpline registration type so that there is some background constraints to match the edges of the head better.",
        argstr="--ROIAutoDilateSize %f",
    )
    ROIAutoClosingSize = traits.Float(
        desc="This flag is only relevant when using ROIAUTO mode for initializing masks.  It defines the hole closing size in mm.  It is rounded up to the nearest whole pixel size in each direction. The default is to use a closing size of 9mm.  For mouse data this value may need to be reset to 0.9 or smaller.",
        argstr="--ROIAutoClosingSize %f",
    )
    numberOfSamples = traits.Int(
        desc="The number of voxels sampled for mutual information computation.  Increase this for higher accuracy, at the cost of longer computation time., NOTE that it is suggested to use samplingPercentage instead of this option. However, if set to non-zero, numberOfSamples overwrites the samplingPercentage option.  ",
        argstr="--numberOfSamples %d",
    )
    strippedOutputTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Rigid component of the estimated affine transform. Can be used to rigidly register the moving image to the fixed image. NOTE:  This value is overridden if either bsplineTransform or linearTransform is set.",
        argstr="--strippedOutputTransform %s",
    )
    transformType = InputMultiPath(
        traits.Str,
        desc="Specifies a list of registration types to be used.  The valid types are, Rigid, ScaleVersor3D, ScaleSkewVersor3D, Affine, BSpline and SyN.  Specifying more than one in a comma separated list will initialize the next stage with the previous results. If registrationClass flag is used, it overrides this parameter setting.",
        sep=",",
        argstr="--transformType %s",
    )
    outputTransform = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="(optional) Filename to which save the (optional) estimated transform. NOTE: You must select either the outputTransform or the outputVolume option.",
        argstr="--outputTransform %s",
    )
    initializeRegistrationByCurrentGenericTransform = traits.Bool(
        desc="If this flag is ON, the current generic composite transform, resulted from the linear registration stages, is set to initialize the follow nonlinear registration process. However, by the default behaviour, the moving image is first warped based on the existant transform before it is passed to the BSpline registration filter. It is done to speed up the BSpline registration by reducing the computations of composite transform Jacobian.",
        argstr="--initializeRegistrationByCurrentGenericTransform ",
    )
    failureExitCode = traits.Int(
        desc="If the fit fails, exit with this status code.  (It can be used to force a successfult exit status of (0) if the registration fails due to reaching the maximum number of iterations.",
        argstr="--failureExitCode %d",
    )
    writeTransformOnFailure = traits.Bool(
        desc="Flag to save the final transform even if the numberOfIterations are reached without convergence. (Intended for use when --failureExitCode 0 )",
        argstr="--writeTransformOnFailure ",
    )
    numberOfThreads = traits.Int(
        desc="Explicitly specify the maximum number of threads to use. (default is auto-detected)",
        argstr="--numberOfThreads %d",
    )
    debugLevel = traits.Int(
        desc="Display debug messages, and produce debug intermediate results.  0=OFF, 1=Minimal, 10=Maximum debugging.",
        argstr="--debugLevel %d",
    )
    costFunctionConvergenceFactor = traits.Float(
        desc="From itkLBFGSBOptimizer.h: Set/Get the CostFunctionConvergenceFactor. Algorithm terminates when the reduction in cost function is less than (factor * epsmcj) where epsmch is the machine precision. Typical values for factor: 1e+12 for low accuracy; 1e+7 for moderate accuracy and 1e+1 for extremely high accuracy.  1e+9 seems to work well.,       ",
        argstr="--costFunctionConvergenceFactor %f",
    )
    projectedGradientTolerance = traits.Float(
        desc="From itkLBFGSBOptimizer.h: Set/Get the ProjectedGradientTolerance. Algorithm terminates when the project gradient is below the tolerance. Default lbfgsb value is 1e-5, but 1e-4 seems to work well.,       ",
        argstr="--projectedGradientTolerance %f",
    )
    maximumNumberOfEvaluations = traits.Int(
        desc="Maximum number of evaluations for line search in lbfgsb optimizer.",
        argstr="--maximumNumberOfEvaluations %d",
    )
    maximumNumberOfCorrections = traits.Int(
        desc="Maximum number of corrections in lbfgsb optimizer.",
        argstr="--maximumNumberOfCorrections %d",
    )
    gui = traits.Bool(
        desc="Display intermediate image volumes for debugging.  NOTE:  This is not part of the standard build sytem, and probably does nothing on your installation.",
        argstr="--gui ",
    )
    promptUser = traits.Bool(
        desc="Prompt the user to hit enter each time an image is sent to the DebugImageViewer",
        argstr="--promptUser ",
    )
    metricSamplingStrategy = traits.Enum(
        "Random",
        desc="It defines the method that registration filter uses to sample the input fixed image. Only Random is supported for now.",
        argstr="--metricSamplingStrategy %s",
    )
    logFileReport = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="A file to write out final information report in CSV file: MetricName,MetricValue,FixedImageName,FixedMaskName,MovingImageName,MovingMaskName",
        argstr="--logFileReport %s",
    )
    writeOutputTransformInFloat = traits.Bool(
        desc="By default, the output registration transforms (either the output composite transform or each transform component) are written to the disk in double precision. If this flag is ON, the output transforms will be written in single (float) precision. It is especially important if the output transform is a displacement field transform, or it is a composite transform that includes several displacement fields.",
        argstr="--writeOutputTransformInFloat ",
    )


class BRAINSFitOutputSpec(TraitedSpec):
    linearTransform = File(
        desc="(optional) Output estimated transform - in case the computed transform is not BSpline. NOTE: You must set at least one output object (transform and/or output volume).",
        exists=True,
    )
    bsplineTransform = File(
        desc="(optional) Output estimated transform - in case the computed transform is BSpline. NOTE: You must set at least one output object (transform and/or output volume).",
        exists=True,
    )
    outputVolume = File(
        desc="(optional) Output image: the moving image warped to the fixed image space. NOTE: You must set at least one output object (transform and/or output volume).",
        exists=True,
    )
    outputFixedVolumeROI = File(
        desc="ROI that is automatically computed from the fixed image. Only available if Masking Option is ROIAUTO. Image areas where the mask volume has zero value are ignored during the registration.",
        exists=True,
    )
    outputMovingVolumeROI = File(
        desc="ROI that is automatically computed from the moving image. Only available if Masking Option is ROIAUTO. Image areas where the mask volume has zero value are ignored during the registration.",
        exists=True,
    )
    strippedOutputTransform = File(
        desc="Rigid component of the estimated affine transform. Can be used to rigidly register the moving image to the fixed image. NOTE:  This value is overridden if either bsplineTransform or linearTransform is set.",
        exists=True,
    )
    outputTransform = File(
        desc="(optional) Filename to which save the (optional) estimated transform. NOTE: You must select either the outputTransform or the outputVolume option.",
        exists=True,
    )
    logFileReport = File(
        desc="A file to write out final information report in CSV file: MetricName,MetricValue,FixedImageName,FixedMaskName,MovingImageName,MovingMaskName",
        exists=True,
    )


class BRAINSFit(SEMLikeCommandLine):
    """title: General Registration (BRAINS)

category: Registration

description: Register a three-dimensional volume to a reference volume (Mattes Mutual Information by default). Full documentation avalable here: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/BRAINSFit. Method described in BRAINSFit: Mutual Information Registrations of Whole-Brain 3D Images, Using the Insight Toolkit, Johnson H.J., Harris G., Williams K., The Insight Journal, 2007. http://hdl.handle.net/1926/1291

version: 3.0.0

documentation-url: http://www.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/BRAINSFit

license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

contributor: Hans J. Johnson, hans-johnson -at- uiowa.edu, http://www.psychiatry.uiowa.edu

acknowledgements: Hans Johnson(1,3,4); Kent Williams(1); Gregory Harris(1), Vincent Magnotta(1,2,3);  Andriy Fedorov(5) 1=University of Iowa Department of Psychiatry, 2=University of Iowa Department of Radiology, 3=University of Iowa Department of Biomedical Engineering, 4=University of Iowa Department of Electrical and Computer Engineering, 5=Surgical Planning Lab, Harvard

"""

    input_spec = BRAINSFitInputSpec
    output_spec = BRAINSFitOutputSpec
    _cmd = " BRAINSFit "
    _outputs_filenames = {
        "outputVolume": "outputVolume.nii",
        "bsplineTransform": "bsplineTransform.h5",
        "outputTransform": "outputTransform.h5",
        "outputFixedVolumeROI": "outputFixedVolumeROI.nii",
        "strippedOutputTransform": "strippedOutputTransform.h5",
        "outputMovingVolumeROI": "outputMovingVolumeROI.nii",
        "linearTransform": "linearTransform.h5",
        "logFileReport": "logFileReport",
    }
    _redirect_x = False



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..resampling import WarpTimeSeriesImageMultiTransform


def test_WarpTimeSeriesImageMultiTransform_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        dimension=dict(
            argstr="%d",
            position=1,
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        input_image=dict(
            argstr="%s",
            copyfile=True,
            extensions=None,
            mandatory=True,
        ),
        invert_affine=dict(),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_postfix=dict(
            argstr="%s",
            usedefault=True,
        ),
        reference_image=dict(
            argstr="-R %s",
            extensions=None,
            xor=["tightest_box"],
        ),
        reslice_by_header=dict(
            argstr="--reslice-by-header",
        ),
        tightest_box=dict(
            argstr="--tightest-bounding-box",
            xor=["reference_image"],
        ),
        transformation_series=dict(
            argstr="%s",
            copyfile=False,
            mandatory=True,
        ),
        use_bspline=dict(
            argstr="--use-Bspline",
        ),
        use_nearest=dict(
            argstr="--use-NN",
        ),
    )
    inputs = WarpTimeSeriesImageMultiTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_WarpTimeSeriesImageMultiTransform_outputs():
    output_map = dict(
        output_image=dict(
            extensions=None,
        ),
    )
    outputs = WarpTimeSeriesImageMultiTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..regutils import RegTransform


def test_RegTransform_inputs():
    input_map = dict(
        aff_2_rig_input=dict(
            argstr="-aff2rig %s",
            extensions=None,
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "flirt_2_nr_input",
            ],
        ),
        args=dict(
            argstr="%s",
        ),
        comp_input=dict(
            argstr="-comp %s",
            extensions=None,
            position=-3,
            requires=["comp_input2"],
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        comp_input2=dict(
            argstr="%s",
            extensions=None,
            position=-2,
        ),
        def_input=dict(
            argstr="-def %s",
            extensions=None,
            position=-2,
            xor=[
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        disp_input=dict(
            argstr="-disp %s",
            extensions=None,
            position=-2,
            xor=[
                "def_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flirt_2_nr_input=dict(
            argstr="-flirtAff2NR %s %s %s",
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
            ],
        ),
        flow_input=dict(
            argstr="-flow %s",
            extensions=None,
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        half_input=dict(
            argstr="-half %s",
            extensions=None,
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        inv_aff_input=dict(
            argstr="-invAff %s",
            extensions=None,
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        inv_nrr_input=dict(
            argstr="-invNrr %s %s",
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        make_aff_input=dict(
            argstr="-makeAff %f %f %f %f %f %f %f %f %f %f %f %f",
            position=-2,
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "upd_s_form_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        omp_core_val=dict(
            argstr="-omp %i",
            usedefault=True,
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        ref1_file=dict(
            argstr="-ref %s",
            extensions=None,
            position=0,
        ),
        ref2_file=dict(
            argstr="-ref2 %s",
            extensions=None,
            position=1,
            requires=["ref1_file"],
        ),
        upd_s_form_input=dict(
            argstr="-updSform %s",
            extensions=None,
            position=-3,
            requires=["upd_s_form_input2"],
            xor=[
                "def_input",
                "disp_input",
                "flow_input",
                "comp_input",
                "inv_aff_input",
                "inv_nrr_input",
                "half_input",
                "make_aff_input",
                "aff_2_rig_input",
                "flirt_2_nr_input",
            ],
        ),
        upd_s_form_input2=dict(
            argstr="%s",
            extensions=None,
            position=-2,
            requires=["upd_s_form_input"],
        ),
    )
    inputs = RegTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_RegTransform_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = RegTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

import os

from ....base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)


class fiberstatsInputSpec(CommandLineInputSpec):
    fiber_file = File(desc="DTI Fiber File", exists=True, argstr="--fiber_file %s")
    verbose = traits.Bool(desc="produce verbose output", argstr="--verbose ")


class fiberstatsOutputSpec(TraitedSpec):
    pass


class fiberstats(SEMLikeCommandLine):
    """title: FiberStats (DTIProcess)

category: Diffusion.Tractography.CommandLineOnly

description: Obsolete tool - Not used anymore

version: 1.1.0

documentation-url: http://www.slicer.org/slicerWiki/index.php/Documentation/Nightly/Extensions/DTIProcess

license: Copyright (c)  Casey Goodlett. All rights reserved.
  See http://www.ia.unc.edu/dev/Copyright.htm for details.
     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

contributor: Casey Goodlett

acknowledgements: Hans Johnson(1,3,4); Kent Williams(1); (1=University of Iowa Department of Psychiatry, 3=University of Iowa Department of Biomedical Engineering, 4=University of Iowa Department of Electrical and Computer Engineering) provided conversions to make DTIProcess compatible with Slicer execution, and simplified the stand-alone build requirements by removing the dependancies on boost and a fortran compiler.

"""

    input_spec = fiberstatsInputSpec
    output_spec = fiberstatsOutputSpec
    _cmd = " fiberstats "
    _outputs_filenames = {}
    _redirect_x = False



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-

import os.path as op

from ..base import (
    CommandLineInputSpec,
    CommandLine,
    traits,
    TraitedSpec,
    File,
    isdefined,
    Undefined,
    InputMultiObject,
)
from .base import MRTrix3BaseInputSpec, MRTrix3Base


class DWIDenoiseInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", position=-2, mandatory=True, desc="input DWI image"
    )
    mask = File(exists=True, argstr="-mask %s", position=1, desc="mask image")
    extent = traits.Tuple(
        (traits.Int, traits.Int, traits.Int),
        argstr="-extent %d,%d,%d",
        desc="set the window size of the denoising filter. (default = 5,5,5)",
    )
    noise = File(
        argstr="-noise %s",
        name_template="%s_noise",
        name_source="in_file",
        keep_extension=True,
        desc="the output noise map",
    )
    out_file = File(
        argstr="%s",
        position=-1,
        name_template="%s_denoised",
        name_source="in_file",
        keep_extension=True,
        desc="the output denoised DWI image",
    )


class DWIDenoiseOutputSpec(TraitedSpec):
    noise = File(desc="the output noise map", exists=True)
    out_file = File(desc="the output denoised DWI image", exists=True)


class DWIDenoise(MRTrix3Base):
    """
    Denoise DWI data and estimate the noise level based on the optimal
    threshold for PCA.

    DWI data denoising and noise map estimation by exploiting data redundancy
    in the PCA domain using the prior knowledge that the eigenspectrum of
    random covariance matrices is described by the universal Marchenko Pastur
    distribution.

    Important note: image denoising must be performed as the first step of the
    image processing pipeline. The routine will fail if interpolation or
    smoothing has been applied to the data prior to denoising.

    Note that this function does not correct for non-Gaussian noise biases.

    For more information, see
    <https://mrtrix.readthedocs.io/en/latest/reference/commands/dwidenoise.html>

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> denoise = mrt.DWIDenoise()
    >>> denoise.inputs.in_file = 'dwi.mif'
    >>> denoise.inputs.mask = 'mask.mif'
    >>> denoise.inputs.noise = 'noise.mif'
    >>> denoise.cmdline                               # doctest: +ELLIPSIS
    'dwidenoise -mask mask.mif -noise noise.mif dwi.mif dwi_denoised.mif'
    >>> denoise.run()                                 # doctest: +SKIP
    """

    _cmd = "dwidenoise"
    input_spec = DWIDenoiseInputSpec
    output_spec = DWIDenoiseOutputSpec


class MRDeGibbsInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", position=-2, mandatory=True, desc="input DWI image"
    )
    axes = traits.ListInt(
        default_value=[0, 1],
        usedefault=True,
        sep=",",
        minlen=2,
        maxlen=2,
        argstr="-axes %s",
        desc="indicate the plane in which the data was acquired (axial = 0,1; "
        "coronal = 0,2; sagittal = 1,2",
    )
    nshifts = traits.Int(
        default_value=20,
        usedefault=True,
        argstr="-nshifts %d",
        desc="discretization of subpixel spacing (default = 20)",
    )
    minW = traits.Int(
        default_value=1,
        usedefault=True,
        argstr="-minW %d",
        desc="left border of window used for total variation (TV) computation "
        "(default = 1)",
    )
    maxW = traits.Int(
        default_value=3,
        usedefault=True,
        argstr="-maxW %d",
        desc="right border of window used for total variation (TV) computation "
        "(default = 3)",
    )
    out_file = File(
        name_template="%s_unr",
        name_source="in_file",
        keep_extension=True,
        argstr="%s",
        position=-1,
        desc="the output unringed DWI image",
    )


class MRDeGibbsOutputSpec(TraitedSpec):
    out_file = File(desc="the output unringed DWI image", exists=True)


class MRDeGibbs(MRTrix3Base):
    """
    Remove Gibbs ringing artifacts.

    This application attempts to remove Gibbs ringing artefacts from MRI images
    using the method of local subvoxel-shifts proposed by Kellner et al.

    This command is designed to run on data directly after it has been
    reconstructed by the scanner, before any interpolation of any kind has
    taken place. You should not run this command after any form of motion
    correction (e.g. not after dwipreproc). Similarly, if you intend running
    dwidenoise, you should run this command afterwards, since it has the
    potential to alter the noise structure, which would impact on dwidenoise's
    performance.

    Note that this method is designed to work on images acquired with full
    k-space coverage. Running this method on partial Fourier ('half-scan') data
    may lead to suboptimal and/or biased results, as noted in the original
    reference below. There is currently no means of dealing with this; users
    should exercise caution when using this method on partial Fourier data, and
    inspect its output for any obvious artefacts.

    For more information, see
    <https://mrtrix.readthedocs.io/en/latest/reference/commands/mrdegibbs.html>

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> unring = mrt.MRDeGibbs()
    >>> unring.inputs.in_file = 'dwi.mif'
    >>> unring.cmdline
    'mrdegibbs -axes 0,1 -maxW 3 -minW 1 -nshifts 20 dwi.mif dwi_unr.mif'
    >>> unring.run()                                 # doctest: +SKIP
    """

    _cmd = "mrdegibbs"
    input_spec = MRDeGibbsInputSpec
    output_spec = MRDeGibbsOutputSpec


class DWIBiasCorrectInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True, argstr="%s", position=-2, mandatory=True, desc="input DWI image"
    )
    in_mask = File(argstr="-mask %s", desc="input mask image for bias field estimation")
    use_ants = traits.Bool(
        argstr="ants",
        mandatory=True,
        desc="use ANTS N4 to estimate the inhomogeneity field",
        position=0,
        xor=["use_fsl"],
    )
    use_fsl = traits.Bool(
        argstr="fsl",
        mandatory=True,
        desc="use FSL FAST to estimate the inhomogeneity field",
        position=0,
        xor=["use_ants"],
    )
    bias = File(argstr="-bias %s", desc="bias field")
    out_file = File(
        name_template="%s_biascorr",
        name_source="in_file",
        keep_extension=True,
        argstr="%s",
        position=-1,
        desc="the output bias corrected DWI image",
        genfile=True,
    )


class DWIBiasCorrectOutputSpec(TraitedSpec):
    bias = File(desc="the output bias field", exists=True)
    out_file = File(desc="the output bias corrected DWI image", exists=True)


class DWIBiasCorrect(MRTrix3Base):
    """
    Perform B1 field inhomogeneity correction for a DWI volume series.

    For more information, see
    <https://mrtrix.readthedocs.io/en/latest/reference/scripts/dwibiascorrect.html>

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> bias_correct = mrt.DWIBiasCorrect()
    >>> bias_correct.inputs.in_file = 'dwi.mif'
    >>> bias_correct.inputs.use_ants = True
    >>> bias_correct.cmdline
    'dwibiascorrect ants dwi.mif dwi_biascorr.mif'
    >>> bias_correct.run()                             # doctest: +SKIP
    """

    _cmd = "dwibiascorrect"
    input_spec = DWIBiasCorrectInputSpec
    output_spec = DWIBiasCorrectOutputSpec
    def _format_arg(self, name, trait_spec, value):
        if name in ("use_ants", "use_fsl"):
            ver = self.version
            # Changed in version 3.0, after release candidates
            if ver is not None and (ver[0] < "3" or ver.startswith("3.0_RC")):
                return f"-{trait_spec.argstr}"
        return super()._format_arg(name, trait_spec, value)

class ResponseSDInputSpec(MRTrix3BaseInputSpec):
    algorithm = traits.Enum(
        "msmt_5tt",
        "dhollander",
        "tournier",
        "tax",
        argstr="%s",
        position=1,
        mandatory=True,
        desc="response estimation algorithm (multi-tissue)",
    )
    in_file = File(
        exists=True, argstr="%s", position=-5, mandatory=True, desc="input DWI image"
    )
    mtt_file = File(argstr="%s", position=-4, desc="input 5tt image")
    wm_file = File(
        "wm.txt",
        argstr="%s",
        position=-3,
        usedefault=True,
        desc="output WM response text file",
    )
    gm_file = File(argstr="%s", position=-2, desc="output GM response text file")
    csf_file = File(argstr="%s", position=-1, desc="output CSF response text file")
    in_mask = File(exists=True, argstr="-mask %s", desc="provide initial mask image")
    max_sh = InputMultiObject(
        traits.Int,
        argstr="-lmax %s",
        sep=",",
        desc=(
            "maximum harmonic degree of response function - single value for "
            "single-shell response, list for multi-shell response"
        ),
    )


class ResponseSDOutputSpec(TraitedSpec):
    wm_file = File(argstr="%s", desc="output WM response text file")
    gm_file = File(argstr="%s", desc="output GM response text file")
    csf_file = File(argstr="%s", desc="output CSF response text file")


class ResponseSD(MRTrix3Base):
    """
    Estimate response function(s) for spherical deconvolution using the specified algorithm.

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> resp = mrt.ResponseSD()
    >>> resp.inputs.in_file = 'dwi.mif'
    >>> resp.inputs.algorithm = 'tournier'
    >>> resp.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> resp.cmdline                               # doctest: +ELLIPSIS
    'dwi2response tournier -fslgrad bvecs bvals dwi.mif wm.txt'
    >>> resp.run()                                 # doctest: +SKIP

    # We can also pass in multiple harmonic degrees in the case of multi-shell
    >>> resp.inputs.max_sh = [6,8,10]
    >>> resp.cmdline
    'dwi2response tournier -fslgrad bvecs bvals -lmax 6,8,10 dwi.mif wm.txt'
    """

    _cmd = "dwi2response"
    input_spec = ResponseSDInputSpec
    output_spec = ResponseSDOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["wm_file"] = op.abspath(self.inputs.wm_file)
        if self.inputs.gm_file != Undefined:
            outputs["gm_file"] = op.abspath(self.inputs.gm_file)
        if self.inputs.csf_file != Undefined:
            outputs["csf_file"] = op.abspath(self.inputs.csf_file)
        return outputs


class ACTPrepareFSLInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="input anatomical image",
    )

    out_file = File(
        "act_5tt.mif",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="output file after processing",
    )


class ACTPrepareFSLOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output response file")


class ACTPrepareFSL(CommandLine):
    """
    Generate anatomical information necessary for Anatomically
    Constrained Tractography (ACT).

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> prep = mrt.ACTPrepareFSL()
    >>> prep.inputs.in_file = 'T1.nii.gz'
    >>> prep.cmdline                               # doctest: +ELLIPSIS
    'act_anat_prepare_fsl T1.nii.gz act_5tt.mif'
    >>> prep.run()                                 # doctest: +SKIP
    """

    _cmd = "act_anat_prepare_fsl"
    input_spec = ACTPrepareFSLInputSpec
    output_spec = ACTPrepareFSLOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


class ReplaceFSwithFIRSTInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-4,
        desc="input anatomical image",
    )
    in_t1w = File(
        exists=True, argstr="%s", mandatory=True, position=-3, desc="input T1 image"
    )
    in_config = File(
        exists=True, argstr="%s", position=-2, desc="connectome configuration file"
    )

    out_file = File(
        "aparc+first.mif",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="output file after processing",
    )


class ReplaceFSwithFIRSTOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output response file")


class ReplaceFSwithFIRST(CommandLine):
    """
    Replace deep gray matter structures segmented with FSL FIRST in a
    FreeSurfer parcellation.

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> prep = mrt.ReplaceFSwithFIRST()
    >>> prep.inputs.in_file = 'aparc+aseg.nii'
    >>> prep.inputs.in_t1w = 'T1.nii.gz'
    >>> prep.inputs.in_config = 'mrtrix3_labelconfig.txt'
    >>> prep.cmdline                               # doctest: +ELLIPSIS
    'fs_parc_replace_sgm_first aparc+aseg.nii T1.nii.gz \
mrtrix3_labelconfig.txt aparc+first.mif'
    >>> prep.run()                                 # doctest: +SKIP
    """

    _cmd = "fs_parc_replace_sgm_first"
    input_spec = ReplaceFSwithFIRSTInputSpec
    output_spec = ReplaceFSwithFIRSTOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import WarpUtils


def test_WarpUtils_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="--in=%s",
            extensions=None,
            mandatory=True,
        ),
        knot_space=dict(
            argstr="--knotspace=%d,%d,%d",
        ),
        out_file=dict(
            argstr="--out=%s",
            extensions=None,
            name_source=["in_file"],
            output_name="out_file",
            position=-1,
        ),
        out_format=dict(
            argstr="--outformat=%s",
        ),
        out_jacobian=dict(
            argstr="--jac=%s",
            extensions=None,
        ),
        output_type=dict(),
        reference=dict(
            argstr="--ref=%s",
            extensions=None,
            mandatory=True,
        ),
        warp_resolution=dict(
            argstr="--warpres=%0.4f,%0.4f,%0.4f",
        ),
        with_affine=dict(
            argstr="--withaff",
        ),
        write_jacobian=dict(
            mandatory=True,
            usedefault=True,
        ),
    )
    inputs = WarpUtils.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_WarpUtils_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
        out_jacobian=dict(
            extensions=None,
        ),
    )
    outputs = WarpUtils.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..epi import PrepareFieldmap


def test_PrepareFieldmap_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        delta_TE=dict(
            argstr="%f",
            mandatory=True,
            position=-2,
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_magnitude=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=3,
        ),
        in_phase=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=2,
        ),
        nocheck=dict(
            argstr="--nocheck",
            position=-1,
            usedefault=True,
        ),
        out_fieldmap=dict(
            argstr="%s",
            extensions=None,
            position=4,
        ),
        output_type=dict(),
        scanner=dict(
            argstr="%s",
            position=1,
            usedefault=True,
        ),
    )
    inputs = PrepareFieldmap.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_PrepareFieldmap_outputs():
    output_map = dict(
        out_fieldmap=dict(
            extensions=None,
        ),
    )
    outputs = PrepareFieldmap.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..registration import MultiResolutionAffineRegistration


def test_MultiResolutionAffineRegistration_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fixedImage=dict(
            argstr="%s",
            extensions=None,
            position=-2,
        ),
        fixedImageMask=dict(
            argstr="--fixedImageMask %s",
            extensions=None,
        ),
        fixedImageROI=dict(
            argstr="--fixedImageROI %s",
        ),
        metricTolerance=dict(
            argstr="--metricTolerance %f",
        ),
        movingImage=dict(
            argstr="%s",
            extensions=None,
            position=-1,
        ),
        numIterations=dict(
            argstr="--numIterations %d",
        ),
        numLineIterations=dict(
            argstr="--numLineIterations %d",
        ),
        resampledImage=dict(
            argstr="--resampledImage %s",
            hash_files=False,
        ),
        saveTransform=dict(
            argstr="--saveTransform %s",
            hash_files=False,
        ),
        stepSize=dict(
            argstr="--stepSize %f",
        ),
        stepTolerance=dict(
            argstr="--stepTolerance %f",
        ),
    )
    inputs = MultiResolutionAffineRegistration.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MultiResolutionAffineRegistration_outputs():
    output_map = dict(
        resampledImage=dict(
            extensions=None,
        ),
        saveTransform=dict(
            extensions=None,
        ),
    )
    outputs = MultiResolutionAffineRegistration.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import os
import numpy as np
import nibabel as nb

import pytest

from nipype.interfaces.freesurfer import model, no_freesurfer
import nipype.pipeline.engine as pe


@pytest.mark.skipif(no_freesurfer(), reason="freesurfer is not installed")
def test_concatenate(tmpdir):
    tmpdir.chdir()

    in1 = tmpdir.join("cont1.nii").strpath
    in2 = tmpdir.join("cont2.nii").strpath
    out = "bar.nii"

    data1 = np.zeros((3, 3, 3, 1), dtype=np.float32)
    data2 = np.ones((3, 3, 3, 5), dtype=np.float32)
    out_data = np.concatenate((data1, data2), axis=3)
    mean_data = np.mean(out_data, axis=3)

    nb.Nifti1Image(data1, affine=np.eye(4)).to_filename(in1)
    nb.Nifti1Image(data2, affine=np.eye(4)).to_filename(in2)

    # Test default behavior
    res = model.Concatenate(in_files=[in1, in2]).run()
    assert res.outputs.concatenated_file == tmpdir.join("concat_output.nii.gz").strpath
    assert np.allclose(nb.load("concat_output.nii.gz").get_fdata(), out_data)

    # Test specified concatenated_file
    res = model.Concatenate(in_files=[in1, in2], concatenated_file=out).run()
    assert res.outputs.concatenated_file == tmpdir.join(out).strpath
    assert np.allclose(nb.load(out).get_fdata(), out_data)

    # Test in workflow
    wf = pe.Workflow("test_concatenate", base_dir=tmpdir.strpath)
    concat = pe.Node(
        model.Concatenate(in_files=[in1, in2], concatenated_file=out), name="concat"
    )
    wf.add_nodes([concat])
    wf.run()
    assert np.allclose(
        nb.load(tmpdir.join("test_concatenate", "concat", out).strpath).get_fdata(),
        out_data,
    )

    # Test a simple statistic
    res = model.Concatenate(
        in_files=[in1, in2], concatenated_file=out, stats="mean"
    ).run()
    assert np.allclose(nb.load(out).get_fdata(), mean_data)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..surface import GrayscaleModelMaker


def test_GrayscaleModelMaker_inputs():
    input_map = dict(
        InputVolume=dict(
            argstr="%s",
            extensions=None,
            position=-2,
        ),
        OutputGeometry=dict(
            argstr="%s",
            hash_files=False,
            position=-1,
        ),
        args=dict(
            argstr="%s",
        ),
        decimate=dict(
            argstr="--decimate %f",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        name=dict(
            argstr="--name %s",
        ),
        pointnormals=dict(
            argstr="--pointnormals ",
        ),
        smooth=dict(
            argstr="--smooth %d",
        ),
        splitnormals=dict(
            argstr="--splitnormals ",
        ),
        threshold=dict(
            argstr="--threshold %f",
        ),
    )
    inputs = GrayscaleModelMaker.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_GrayscaleModelMaker_outputs():
    output_map = dict(
        OutputGeometry=dict(
            extensions=None,
            position=-1,
        ),
    )
    outputs = GrayscaleModelMaker.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..dcmstack import DcmStack


def test_DcmStack_inputs():
    input_map = dict(
        dicom_files=dict(
            mandatory=True,
        ),
        embed_meta=dict(),
        exclude_regexes=dict(),
        force_read=dict(
            usedefault=True,
        ),
        include_regexes=dict(),
        out_ext=dict(
            usedefault=True,
        ),
        out_format=dict(),
        out_path=dict(),
    )
    inputs = DcmStack.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DcmStack_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = DcmStack.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import FLAMEO


def test_FLAMEO_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        burnin=dict(
            argstr="--burnin=%d",
        ),
        cope_file=dict(
            argstr="--copefile=%s",
            extensions=None,
            mandatory=True,
        ),
        cov_split_file=dict(
            argstr="--covsplitfile=%s",
            extensions=None,
            mandatory=True,
        ),
        design_file=dict(
            argstr="--designfile=%s",
            extensions=None,
            mandatory=True,
        ),
        dof_var_cope_file=dict(
            argstr="--dofvarcopefile=%s",
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        f_con_file=dict(
            argstr="--fcontrastsfile=%s",
            extensions=None,
        ),
        fix_mean=dict(
            argstr="--fixmean",
        ),
        infer_outliers=dict(
            argstr="--inferoutliers",
        ),
        log_dir=dict(
            argstr="--ld=%s",
            usedefault=True,
        ),
        mask_file=dict(
            argstr="--maskfile=%s",
            extensions=None,
            mandatory=True,
        ),
        n_jumps=dict(
            argstr="--njumps=%d",
        ),
        no_pe_outputs=dict(
            argstr="--nopeoutput",
        ),
        outlier_iter=dict(
            argstr="--ioni=%d",
        ),
        output_type=dict(),
        run_mode=dict(
            argstr="--runmode=%s",
            mandatory=True,
        ),
        sample_every=dict(
            argstr="--sampleevery=%d",
        ),
        sigma_dofs=dict(
            argstr="--sigma_dofs=%d",
        ),
        t_con_file=dict(
            argstr="--tcontrastsfile=%s",
            extensions=None,
            mandatory=True,
        ),
        var_cope_file=dict(
            argstr="--varcopefile=%s",
            extensions=None,
        ),
    )
    inputs = FLAMEO.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FLAMEO_outputs():
    output_map = dict(
        copes=dict(),
        fstats=dict(),
        mrefvars=dict(),
        pes=dict(),
        res4d=dict(),
        stats_dir=dict(),
        tdof=dict(),
        tstats=dict(),
        var_copes=dict(),
        weights=dict(),
        zfstats=dict(),
        zstats=dict(),
    )
    outputs = FLAMEO.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""MRTrix version 2 (DEPRECATED) -- tools to perform various types of diffusion MRI analyses."""
from .tracking import (
    Tracks2Prob,
    FilterTracks,
    StreamlineTrack,
    DiffusionTensorStreamlineTrack,
    SphericallyDeconvolutedStreamlineTrack,
    ProbabilisticSphericallyDeconvolutedStreamlineTrack,
)
from .tensors import (
    FSL2MRTrix,
    ConstrainedSphericalDeconvolution,
    DWI2SphericalHarmonicsImage,
    EstimateResponseForSH,
    GenerateDirections,
    FindShPeaks,
    Directions2Amplitude,
)
from .preprocess import (
    MRConvert,
    MRMultiply,
    MRTrixViewer,
    MRTrixInfo,
    GenerateWhiteMatterMask,
    DWI2Tensor,
    Tensor2ApparentDiffusion,
    Tensor2FractionalAnisotropy,
    Tensor2Vector,
    MedianFilter3D,
    Erode,
    Threshold,
)
from .convert import MRTrix2TrackVis



# -*- coding: utf-8 -*-
"""MIPAV enables quantitative analysis and visualization of multimodal medical images."""
from .developer import (
    JistLaminarVolumetricLayering,
    JistBrainMgdmSegmentation,
    JistLaminarProfileGeometry,
    JistLaminarProfileCalculator,
    MedicAlgorithmN3,
    JistLaminarROIAveraging,
    MedicAlgorithmLesionToads,
    JistBrainMp2rageSkullStripping,
    JistCortexSurfaceMeshInflation,
    RandomVol,
    MedicAlgorithmImageCalculator,
    JistBrainMp2rageDuraEstimation,
    JistLaminarProfileSampling,
    MedicAlgorithmMipavReorient,
    MedicAlgorithmSPECTRE2010,
    JistBrainPartialVolumeFilter,
    JistIntensityMp2rageMasking,
    MedicAlgorithmThresholdToBinaryMask,
)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Miscellaneous algorithms for 2D contours and 3D triangularized meshes handling
"""
import os.path as op
import numpy as np
from numpy import linalg as nla

from .. import logging
from ..interfaces.base import (
    BaseInterface,
    traits,
    TraitedSpec,
    File,
    BaseInterfaceInputSpec,
)
from ..interfaces.vtkbase import tvtk
from ..interfaces import vtkbase as VTKInfo

IFLOGGER = logging.getLogger("nipype.interface")


class TVTKBaseInterface(BaseInterface):
    """ A base class for interfaces using VTK """

    _redirect_x = True

    def __init__(self, **inputs):
        if VTKInfo.no_tvtk():
            raise ImportError("This interface requires tvtk to run.")
        super(TVTKBaseInterface, self).__init__(**inputs)


class WarpPointsInputSpec(BaseInterfaceInputSpec):
    points = File(exists=True, mandatory=True, desc="file containing the point set")
    warp = File(
        exists=True, mandatory=True, desc="dense deformation field to be applied"
    )
    interp = traits.Enum(
        "cubic",
        "nearest",
        "linear",
        usedefault=True,
        mandatory=True,
        desc="interpolation",
    )
    out_points = File(
        name_source="points",
        name_template="%s_warped",
        output_name="out_points",
        keep_extension=True,
        desc="the warped point set",
    )


class WarpPointsOutputSpec(TraitedSpec):
    out_points = File(desc="the warped point set")


class WarpPoints(TVTKBaseInterface):
    """
    Applies a displacement field to a point set given in vtk format.
    Any discrete deformation field, given in physical coordinates and
    which volume covers the extent of the vtk point set, is a valid
    ``warp`` file. FSL interfaces are compatible, for instance any
    field computed with :class:`nipype.interfaces.fsl.utils.ConvertWarp`.

    Example::

        from nipype.algorithms.mesh import WarpPoints
        wp = WarpPoints()
        wp.inputs.points = 'surf1.vtk'
        wp.inputs.warp = 'warpfield.nii'
        res = wp.run()

    """

    input_spec = WarpPointsInputSpec
    output_spec = WarpPointsOutputSpec

    def _gen_fname(self, in_file, suffix="generated", ext=None):
        fname, fext = op.splitext(op.basename(in_file))

        if fext == ".gz":
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext

        if ext is None:
            ext = fext

        if ext[0] == ".":
            ext = ext[1:]
        return op.abspath("%s_%s.%s" % (fname, suffix, ext))

    def _run_interface(self, runtime):
        import nibabel as nb
        from scipy import ndimage

        r = tvtk.PolyDataReader(file_name=self.inputs.points)
        r.update()
        mesh = VTKInfo.vtk_output(r)
        points = np.array(mesh.points)
        warp_dims = nb.funcs.four_to_three(nb.load(self.inputs.warp))

        affine = warp_dims[0].affine
        # voxsize = warp_dims[0].header.get_zooms()
        vox2ras = affine[0:3, 0:3]
        ras2vox = np.linalg.inv(vox2ras)
        origin = affine[0:3, 3]
        voxpoints = np.array([np.dot(ras2vox, (p - origin)) for p in points])

        warps = []
        for axis in warp_dims:
            wdata = axis.dataobj  # four_to_three ensures this is an array
            if np.any(wdata != 0):
                warp = ndimage.map_coordinates(wdata, voxpoints.transpose())
            else:
                warp = np.zeros((points.shape[0],))

            warps.append(warp)

        disps = np.squeeze(np.dstack(warps))
        newpoints = [p + d for p, d in zip(points, disps)]
        mesh.points = newpoints
        w = tvtk.PolyDataWriter()
        VTKInfo.configure_input_data(w, mesh)
        w.file_name = self._gen_fname(self.inputs.points, suffix="warped", ext=".vtk")
        w.write()
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_points"] = self._gen_fname(
            self.inputs.points, suffix="warped", ext=".vtk"
        )
        return outputs


class ComputeMeshWarpInputSpec(BaseInterfaceInputSpec):
    surface1 = File(
        exists=True,
        mandatory=True,
        desc=("Reference surface (vtk format) to which compute " "distance."),
    )
    surface2 = File(
        exists=True,
        mandatory=True,
        desc=("Test surface (vtk format) from which compute " "distance."),
    )
    metric = traits.Enum(
        "euclidean", "sqeuclidean", usedefault=True, desc="norm used to report distance"
    )
    weighting = traits.Enum(
        "none",
        "area",
        usedefault=True,
        desc=(
            '"none": no weighting is performed, surface": edge distance is '
            "weighted by the corresponding surface area"
        ),
    )
    out_warp = File(
        "surfwarp.vtk",
        usedefault=True,
        desc="vtk file based on surface1 and warpings mapping it " "to surface2",
    )
    out_file = File(
        "distance.npy",
        usedefault=True,
        desc="numpy file keeping computed distances and weights",
    )


class ComputeMeshWarpOutputSpec(TraitedSpec):
    distance = traits.Float(desc="computed distance")
    out_warp = File(
        exists=True,
        desc=("vtk file with the vertex-wise " "mapping of surface1 to surface2"),
    )
    out_file = File(
        exists=True, desc="numpy file keeping computed distances and weights"
    )


class ComputeMeshWarp(TVTKBaseInterface):
    """
    Calculates a the vertex-wise warping to get surface2 from surface1.
    It also reports the average distance of vertices, using the norm specified
    as input.

    .. warning:

      A point-to-point correspondence between surfaces is required


    Example::

        import nipype.algorithms.mesh as m
        dist = m.ComputeMeshWarp()
        dist.inputs.surface1 = 'surf1.vtk'
        dist.inputs.surface2 = 'surf2.vtk'
        res = dist.run()

    """

    input_spec = ComputeMeshWarpInputSpec
    output_spec = ComputeMeshWarpOutputSpec

    def _triangle_area(self, A, B, C):
        A = np.array(A)
        B = np.array(B)
        C = np.array(C)
        ABxAC = nla.norm(A - B) * nla.norm(A - C)
        prod = np.dot(B - A, C - A)
        angle = np.arccos(prod / ABxAC)
        area = 0.5 * ABxAC * np.sin(angle)
        return area

    def _run_interface(self, runtime):
        r1 = tvtk.PolyDataReader(file_name=self.inputs.surface1)
        r2 = tvtk.PolyDataReader(file_name=self.inputs.surface2)
        vtk1 = VTKInfo.vtk_output(r1)
        vtk2 = VTKInfo.vtk_output(r2)
        r1.update()
        r2.update()
        assert len(vtk1.points) == len(vtk2.points)

        points1 = np.array(vtk1.points)
        points2 = np.array(vtk2.points)

        diff = points2 - points1
        weights = np.ones(len(diff))

        try:
            errvector = nla.norm(diff, axis=1)
        except TypeError:  # numpy < 1.9
            errvector = np.apply_along_axis(nla.norm, 1, diff)

        if self.inputs.metric == "sqeuclidean":
            errvector **= 2

        if self.inputs.weighting == "area":
            faces = vtk1.polys.to_array().reshape(-1, 4).astype(int)[:, 1:]

            for i, p1 in enumerate(points2):
                # compute surfaces, set in weight
                w = 0.0
                point_faces = faces[(faces[:, :] == i).any(axis=1)]

                for idset in point_faces:
                    fp1 = points1[int(idset[0])]
                    fp2 = points1[int(idset[1])]
                    fp3 = points1[int(idset[2])]
                    w += self._triangle_area(fp1, fp2, fp3)
                weights[i] = w

        result = np.vstack([errvector, weights])
        np.save(op.abspath(self.inputs.out_file), result.transpose())

        out_mesh = tvtk.PolyData()
        out_mesh.points = vtk1.points
        out_mesh.polys = vtk1.polys
        out_mesh.point_data.vectors = diff
        out_mesh.point_data.vectors.name = "warpings"
        writer = tvtk.PolyDataWriter(file_name=op.abspath(self.inputs.out_warp))
        VTKInfo.configure_input_data(writer, out_mesh)
        writer.write()

        self._distance = np.average(errvector, weights=weights)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        outputs["out_warp"] = op.abspath(self.inputs.out_warp)
        outputs["distance"] = self._distance
        return outputs


class MeshWarpMathsInputSpec(BaseInterfaceInputSpec):
    in_surf = File(
        exists=True,
        mandatory=True,
        desc=(
            "Input surface in vtk format, with associated warp "
            "field as point data (ie. from ComputeMeshWarp"
        ),
    )
    float_trait = traits.Either(
        traits.Float(1.0),
        traits.Tuple(traits.Float(1.0), traits.Float(1.0), traits.Float(1.0)),
    )

    operator = traits.Either(
        float_trait,
        File(exists=True),
        default=1.0,
        usedefault=True,
        mandatory=True,
        desc="image, float or tuple of floats to act as operator",
    )

    operation = traits.Enum(
        "sum", "sub", "mul", "div", usedefault=True, desc="operation to be performed"
    )

    out_warp = File(
        "warp_maths.vtk",
        usedefault=True,
        desc="vtk file based on in_surf and warpings mapping it " "to out_file",
    )
    out_file = File("warped_surf.vtk", usedefault=True, desc="vtk with surface warped")


class MeshWarpMathsOutputSpec(TraitedSpec):
    out_warp = File(
        exists=True,
        desc=("vtk file with the vertex-wise " "mapping of surface1 to surface2"),
    )
    out_file = File(exists=True, desc="vtk with surface warped")


class MeshWarpMaths(TVTKBaseInterface):
    """
    Performs the most basic mathematical operations on the warping field
    defined at each vertex of the input surface. A surface with scalar
    or vector data can be used as operator for non-uniform operations.

    .. warning:

      A point-to-point correspondence between surfaces is required


    Example::

        import nipype.algorithms.mesh as m
        mmath = m.MeshWarpMaths()
        mmath.inputs.in_surf = 'surf1.vtk'
        mmath.inputs.operator = 'surf2.vtk'
        mmath.inputs.operation = 'mul'
        res = mmath.run()

    """

    input_spec = MeshWarpMathsInputSpec
    output_spec = MeshWarpMathsOutputSpec

    def _run_interface(self, runtime):
        r1 = tvtk.PolyDataReader(file_name=self.inputs.in_surf)
        vtk1 = VTKInfo.vtk_output(r1)
        r1.update()
        points1 = np.array(vtk1.points)

        if vtk1.point_data.vectors is None:
            raise RuntimeError("No warping field was found in in_surf")

        operator = self.inputs.operator
        opfield = np.ones_like(points1)

        if isinstance(operator, (str, bytes)):
            r2 = tvtk.PolyDataReader(file_name=self.inputs.surface2)
            vtk2 = VTKInfo.vtk_output(r2)
            r2.update()
            assert len(points1) == len(vtk2.points)

            opfield = vtk2.point_data.vectors

            if opfield is None:
                opfield = vtk2.point_data.scalars

            if opfield is None:
                raise RuntimeError("No operator values found in operator file")

            opfield = np.array(opfield)

            if opfield.shape[1] < points1.shape[1]:
                opfield = np.array([opfield.tolist()] * points1.shape[1]).T
        else:
            operator = np.atleast_1d(operator)
            opfield *= operator

        warping = np.array(vtk1.point_data.vectors)

        if self.inputs.operation == "sum":
            warping += opfield
        elif self.inputs.operation == "sub":
            warping -= opfield
        elif self.inputs.operation == "mul":
            warping *= opfield
        elif self.inputs.operation == "div":
            warping /= opfield

        vtk1.point_data.vectors = warping
        writer = tvtk.PolyDataWriter(file_name=op.abspath(self.inputs.out_warp))
        VTKInfo.configure_input_data(writer, vtk1)
        writer.write()

        vtk1.point_data.vectors = None
        vtk1.points = points1 + warping
        writer = tvtk.PolyDataWriter(file_name=op.abspath(self.inputs.out_file))
        VTKInfo.configure_input_data(writer, vtk1)
        writer.write()
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        outputs["out_warp"] = op.abspath(self.inputs.out_warp)
        return outputs


class P2PDistance(ComputeMeshWarp):
    """
    Calculates a point-to-point (p2p) distance between two corresponding
    VTK-readable meshes or contours.

    A point-to-point correspondence between nodes is required

    .. deprecated:: 1.0-dev
       Use :py:class:`ComputeMeshWarp` instead.
    """

    def __init__(self, **inputs):
        super(P2PDistance, self).__init__(**inputs)
        IFLOGGER.warning(
            "This interface has been deprecated since 1.0, please "
            "use ComputeMeshWarp"
        )



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..odf import SFPeaks


def test_SFPeaks_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        density=dict(
            argstr="-density %d",
            units="NA",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-inputfile %s",
            extensions=None,
            mandatory=True,
        ),
        inputmodel=dict(
            argstr="-inputmodel %s",
            mandatory=True,
        ),
        mepointset=dict(
            argstr="-mepointset %d",
            units="NA",
        ),
        noconsistencycheck=dict(
            argstr="-noconsistencycheck",
        ),
        numpds=dict(
            argstr="-numpds %d",
            units="NA",
        ),
        order=dict(
            argstr="-order %d",
            units="NA",
        ),
        out_file=dict(
            argstr="> %s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        pdthresh=dict(
            argstr="-pdthresh %f",
            units="NA",
        ),
        pointset=dict(
            argstr="-pointset %d",
            units="NA",
        ),
        rbfpointset=dict(
            argstr="-rbfpointset %d",
            units="NA",
        ),
        scheme_file=dict(
            argstr="%s",
            extensions=None,
        ),
        searchradius=dict(
            argstr="-searchradius %f",
            units="NA",
        ),
        stdsfrommean=dict(
            argstr="-stdsfrommean %f",
            units="NA",
        ),
    )
    inputs = SFPeaks.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_SFPeaks_outputs():
    output_map = dict(
        peaks=dict(
            extensions=None,
        ),
    )
    outputs = SFPeaks.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from .specialized import (
    BRAINSCut,
    BRAINSROIAuto,
    BRAINSConstellationDetector,
    BRAINSCreateLabelMapFromProbabilityMaps,
    BinaryMaskEditorBasedOnLandmarks,
    BRAINSMultiSTAPLE,
    BRAINSABC,
    ESLR,
)
